import abc

import pydantic
import pydantic_core
from typing_extensions import Any, Literal, Self, TypeVar, override

from svglab import protocols


_T_contra = TypeVar("_T_contra", contravariant=True)
_SupportsNegT_contra = TypeVar(
    "_SupportsNegT_contra", bound=protocols.SupportsNeg, contravariant=True
)
_SupportsRTrueDivT_contra = TypeVar(
    "_SupportsRTrueDivT_contra",
    bound=protocols.SupportsRTrueDiv[Any],
    contravariant=True,
)


class CustomModel(protocols.PydanticCompatible, metaclass=abc.ABCMeta):
    """A mixin for easy creation of custom pydantic-compatible classes.

    This class is a mixin for classes that need to be pydantic-compatible,
    but are not, for technical reasons, pydantic dataclasses or subclasses
    of `BaseModel` (for example collections).

    By implementing the `_validate` class method, the class can be used
    in pydantic models. `__get_pydantic_core_schema__` is created
    automatically and should not be overridden.
    """

    @classmethod
    @abc.abstractmethod
    def _validate(
        cls, value: object, info: pydantic_core.core_schema.ValidationInfo
    ) -> Self:
        """Validate the value and return a new instance of the class."""

    @override
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: type, handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.core_schema.CoreSchema:
        del source_type, handler

        return (
            pydantic_core.core_schema.with_info_plain_validator_function(
                function=cls._validate
            )
        )


class SubWithNeg(
    protocols.SupportsAdd[_SupportsNegT_contra],
    protocols.SupportsSub[_SupportsNegT_contra],
    metaclass=abc.ABCMeta,
):
    @override
    def __sub__(self, other: _SupportsNegT_contra, /) -> Self:
        return self + -other


class RAddWithAdd(
    protocols.SupportsRAdd[_T_contra],
    protocols.SupportsAdd[_T_contra],
    metaclass=abc.ABCMeta,
):
    @override
    def __radd__(self, other: _T_contra, /) -> Self:
        return self + other


class RSubWithSub(
    protocols.SupportsRSub[_T_contra],
    protocols.SupportsSub[_T_contra],
    metaclass=abc.ABCMeta,
):
    @override
    def __rsub__(self, other: _T_contra, /) -> Self:
        return self - other


class RMulWithMul(
    protocols.SupportsRMul[_T_contra],
    protocols.SupportsMul[_T_contra],
    metaclass=abc.ABCMeta,
):
    @override
    def __rmul__(self, other: _T_contra, /) -> Self:
        return self * other


class RTruedivWithTrueDiv(
    protocols.SupportsRTrueDiv[_T_contra],
    protocols.SupportsTrueDiv[_T_contra],
    metaclass=abc.ABCMeta,
):
    @override
    def __rtruediv__(self, other: _T_contra, /) -> Self:
        return self / other


class NegWithMul(
    protocols.SupportsNeg,
    protocols.SupportsMul[Literal[-1]],
    metaclass=abc.ABCMeta,
):
    @override
    def __neg__(self) -> Self:
        return self * -1


class TrueDivWithMul(
    protocols.SupportsTrueDiv[_SupportsRTrueDivT_contra],
    protocols.SupportsMul[_SupportsRTrueDivT_contra],
    metaclass=abc.ABCMeta,
):
    @override
    def __truediv__(self, other: _SupportsRTrueDivT_contra, /) -> Self:
        return self * (1 / other)


class FloatMulDiv(
    NegWithMul,
    RMulWithMul[float],
    RTruedivWithTrueDiv[float],
    TrueDivWithMul[float],
    metaclass=abc.ABCMeta,
):
    pass


class AddSub(
    SubWithNeg[_SupportsNegT_contra],
    RAddWithAdd[_SupportsNegT_contra],
    RSubWithSub[_SupportsNegT_contra],
    metaclass=abc.ABCMeta,
):
    pass



import copy
import io
import uuid

import numpy as np
import numpy.typing as npt
import PIL.Image
import resvg_py
from typing_extensions import (
    Final,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
    runtime_checkable,
)

from svglab import errors, serialize, utils
from svglab.attrparse import color, length
from svglab.elements import common


Mask: TypeAlias = npt.NDArray[np.bool_]
BBox: TypeAlias = tuple[int, int, int, int]

_ImageArray: TypeAlias = npt.NDArray[np.uint8]

_TagT = TypeVar("_TagT", bound=common.Tag)

_BLACK: Final = color.Color((0, 0, 0))


@runtime_checkable
class _SvgTagLike(Protocol):
    width: length.Length | None
    height: length.Length | None

    def render(
        self, width: float | None = None, height: float | None = None
    ) -> PIL.Image.Image: ...


def _length_to_user_units(length: length.Length | None) -> float | None:
    """Convert a length to user units, if possible.

    Args:
        length: The length to convert, or `None`.

    Returns:
        The length in user units, or `None` if the length cannot be converted.

    """
    if length is None:
        return None

    try:
        return float(length)
    except errors.SvgUnitConversionError:
        return None


def render(
    svg: common.Tag,
    *,
    width: float | None = None,
    height: float | None = None,
) -> PIL.Image.Image:
    if not isinstance(svg, _SvgTagLike):
        raise TypeError("Tag must be an SVG element")

    svg_width = _length_to_user_units(svg.width)
    svg_height = _length_to_user_units(svg.height)

    if (
        width is not None
        and height is not None
        and svg_width is not None
        and svg_height is not None
        and width / height != svg_width / svg_height
    ):
        msg = (
            "Aspect ratio mismatch: "
            f"{svg.width=}, {svg.height=}, {width=}, {height=}"
        )
        raise ValueError(msg)

    if svg_width is not None and svg_height is not None:
        if width is not None and height is None:
            ratio = width / svg_width
            height = svg_height * ratio

        if width is None and height is not None:
            ratio = height / svg_height
            width = svg_width * ratio

    width = width if width is not None else svg_width
    height = height if height is not None else svg_height

    if width is None or height is None:
        msg = (
            "Unable to determine image dimensions: "
            f"{svg.width=}, {svg.height=}, {width=}, {height=}"
        )
        raise ValueError(msg)

    svg = copy.copy(svg)

    svg.width = length.Length(width)
    svg.height = length.Length(height)

    xml = svg.to_xml(formatter=serialize.MINIMAL_FORMATTER)
    raw = cast(bytes, resvg_py.svg_to_bytes(svg_string=xml))

    return PIL.Image.open(io.BytesIO(raw))


def _copy_tree(tag: _TagT) -> tuple[_TagT, _SvgTagLike]:
    """Resolve the root `Svg` tag and create a deep copy of the SVG tree.

    The source tag is identified in the copied tree and returned for easy
    access.

    Args:
        tag: The tag in the SVG tree to copy.

    Returns:
        A tuple of the copied tag and the root `Svg` tag of the copied tree.

    Raises:
        ValueError: If the tag is not a part of an SVG tree.

    """
    svg = utils.take_last(tag.parents)

    if not isinstance(svg, _SvgTagLike):
        raise ValueError("Tag must be part of an SVG tree")  # noqa: TRY004

    original_id = tag.id
    tag.id = uuid.uuid4().hex

    try:
        svg = copy.deepcopy(svg)

        candidates = svg.find_all(type(tag))

        this = next(tag for tag in candidates if tag.id == tag.id)
    finally:
        tag.id = original_id

    return this, svg


def _render_tree(
    tag: common.Tag,
    *,
    render_this: bool,
    render_other: bool,
    make_tag_visible: bool,
    width: float | None = None,
    height: float | None = None,
) -> PIL.Image.Image:
    """Resolve the root `Svg` tag and render the SVG tree to an image.

    Args:
        tag: The tag in the SVG tree to render.
        render_this: Whether to render the specified tag.
        render_other: Whether to render all other tags in the tree.
        make_tag_visible: Whether to attempt to make the specified tag visible,
            even if it would normally not be rendered (e.g., if due to a
            transparent fill).
        width: The width of the rendered image, in pixels. If `None`, the width
            attribute of the SVG element is used.
        height: The height of the rendered image, in pixels. If `None`, the
            height attribute of the SVG element is used.

    Returns:
        The rendered image.

    Raises:
        ValueError: If `make_tag_visible` is `True` and `render_this`
        is `False`.

    """
    if make_tag_visible and not render_this:
        raise ValueError(
            "make_tag_visible cannot be True if render_this is False"
        )

    tag_copy, svg = _copy_tree(tag)
    assert isinstance(svg, common.Tag)

    for t in svg.find_all():
        t.visibility = "visible" if render_other else "hidden"

    tag_copy.visibility = "visible" if render_this else "hidden"

    if make_tag_visible:
        del tag_copy.display
        tag_copy.fill = _BLACK
        tag_copy.fill_opacity = 1
        tag_copy.opacity = 1
        tag_copy.stroke = _BLACK
        tag_copy.stroke_opacity = 1
        tag_copy.visibility = "visible"

    return svg.render(width=width, height=height)


def _mask_to_image(mask: Mask) -> PIL.Image.Image:
    """Convert boolean mask into an RGBA image.

    Areas where the mask is True are set to solid black. All other areas are
    fully transparent. This allows convenient use with
    `PIL.Image.Image.getbbox()`.

    Args:
        mask: A boolean mask given as an NDArray of shape (x, y), where x
            and y are the dimensions of the resulting image.

    Returns:
        An RGBA image representing the mask.

    """
    x, y = mask.shape

    # create fully transparent image
    rgba: _ImageArray = np.zeros((x, y, 4), dtype=np.uint8)

    # set to solid black where mask is True
    rgba[mask, 3] = 255

    return PIL.Image.fromarray(rgba)


def mask(
    tag: common.Tag,
    *,
    width: float | None = None,
    height: float | None = None,
) -> Mask:
    img = _render_tree(
        tag,
        render_this=True,
        render_other=False,
        make_tag_visible=True,
        width=width,
        height=height,
    )
    array: Mask = np.array(img)

    return array[:, :, 3] > 0  # alpha channel > 0


def visible_mask(
    tag: common.Tag,
    *,
    width: float | None = None,
    height: float | None = None,
) -> Mask:
    without_tag = _render_tree(
        tag,
        render_this=False,
        render_other=True,
        make_tag_visible=False,
        width=width,
        height=height,
    )

    with_tag = _render_tree(
        tag,
        render_this=True,
        render_other=True,
        make_tag_visible=False,
        width=width,
        height=height,
    )

    without_tag_array: _ImageArray = np.array(without_tag)
    with_tag_array: _ImageArray = np.array(with_tag)

    diff = np.any(without_tag_array != with_tag_array, axis=2)
    assert isinstance(diff, np.ndarray)

    return diff


def bbox(tag: common.Tag) -> BBox | None:
    img = _render_tree(
        tag, render_this=True, render_other=False, make_tag_visible=True
    )

    return img.getbbox()


def visible_bbox(tag: common.Tag) -> BBox | None:
    mask = visible_mask(tag)
    img = _mask_to_image(mask)

    return img.getbbox()



import collections
import functools
import math
import re
from collections.abc import Callable, Generator, Iterable, Sequence, Sized

import bs4
from typing_extensions import (
    Any,
    Literal,
    SupportsFloat,
    SupportsIndex,
    TypeAlias,
    TypeIs,
    TypeVar,
    cast,
    overload,
    override,
)
from useful_types import SupportsRichComparisonT

from svglab import constants, serialize
from svglab.elements import names


_T = TypeVar("_T")
_DT = TypeVar("_DT")
_NT = TypeVar("_NT")

_Map: TypeAlias = Callable[[_T], _NT]


class BsFormatter(bs4.formatter.XMLFormatter):
    def __init__(self) -> None:
        formatter = serialize.get_current_formatter()
        super().__init__(
            # replaces special characters with XML entities (e.g. < -> &lt;)
            entity_substitution=bs4.formatter.EntitySubstitution.substitute_xml,
            indent=formatter.indent,
        )

    @override
    def attributes(self, tag: bs4.Tag) -> Iterable[tuple[str, Any]]:
        formatter = serialize.get_current_formatter()
        order = []

        if tag.name in formatter.attribute_order:
            order = formatter.attribute_order[
                cast(names.TagName, tag.name)
            ]
        elif "*" in formatter.attribute_order:
            order = formatter.attribute_order["*"]

        order_index = {attr: i for i, attr in enumerate(order)}
        order_length = len(order)

        # first sort by the predefined order, then by name
        return sorted(
            tag.attrs.items(),
            key=lambda item: (
                order_index.get(item[0], order_length),
                item[0],
            ),
        )


def take_last(iterable: Iterable[_T], /) -> _T | None:
    """Get the last item in an iterable.

    Args:
        iterable: The iterable to extract the last item from.

    Returns:
        The last item in the iterable, or `None` if the iterable is empty.

    Examples:
        >>> take_last([]) is None
        True
        >>> take_last([1, 2, 3])
        3
        >>> take_last(range(0)) is None
        True
        >>> take_last(range(3))
        2

    """
    return functools.reduce(lambda _, s: s, iterable, None)


def _make_soup(element: bs4.PageElement, /) -> bs4.BeautifulSoup:
    soup = bs4.BeautifulSoup()
    soup.append(element)
    return soup


def beautifulsoup_to_str(
    element: bs4.PageElement, /, *, pretty: bool
) -> str:
    result: str

    match element, pretty:
        case bs4.NavigableString(), _:
            result = _make_soup(element).decode(formatter=BsFormatter())
        case bs4.Tag(), True:
            soup = _make_soup(element)

            result = soup.prettify(formatter=BsFormatter())
        case bs4.Tag(), False:
            result = str(element)
        case _:
            msg = f"Unsupported type: {type(element)}"
            raise TypeError(msg)

    return result.strip()


def clamp(
    value: SupportsRichComparisonT,
    /,
    *,
    min_value: SupportsRichComparisonT,
    max_value: SupportsRichComparisonT,
) -> SupportsRichComparisonT:
    """Clamp a value between two bounds.

    Args:
        value: The value to clamp.
        min_value: The minimum value.
        max_value: The maximum value.

    Returns:
        The clamped value. If `value` is less than `min_value`, `min_value` is
        returned. If `value` is greater than `max_value`, `max_value` is
        returned. Otherwise, `value` is returned.

    Examples:
        >>> clamp(5, min_value=0, max_value=10)
        5
        >>> clamp(-5, min_value=0, max_value=10)
        0
        >>> clamp(15, min_value=0, max_value=10)
        10

    """
    return max(min(value, max_value), min_value)


def get_all_subclasses(cls: type[_T], /) -> Generator[type[_T]]:
    """Recursively obtain all subclasses of a class.

    Args:
        cls: The class to obtain subclasses for.

    Yields:
        Subclasses of the given class, including subclasses of subclasses
        (and so on).

    Examples:
        >>> class A:
        ...     pass
        >>> class B(A):
        ...     pass
        >>> class C(B):
        ...     pass
        >>> list(cls.__name__ for cls in get_all_subclasses(A))
        ['B', 'C']

    """
    queue = collections.deque([cls])

    while queue:
        subclass = queue.popleft()

        if subclass is not cls:
            yield subclass

        queue.extend(subclass.__subclasses__())


def basic_compare(other: object, /, *, self: _T) -> TypeIs[_T]:
    """Perform a basic comparison between two objects.

    This is a helper function for implementing the `__eq__` method.

    It checks if the other object is the same as the current object or
    an instance of the same class.

    Args:
        other: The other object to compare.
        self: The current object.

    Returns:
        `True` if the other object is the same as the current object or an
        instance of the same class, `False` otherwise.

    Examples:
        >>> class A:
        ...     def __eq__(self, other: object, /) -> bool:
        ...         return basic_compare(other, self=self)
        >>> A() == A()
        True
        >>> A() == 1
        False

    """
    return other is self or isinstance(other, type(self))


def prev(sequence: Sequence[_T], item: _T) -> _T:
    """Get the item before a given item in a sequence.

    Args:
        sequence: The sequence to search.
        item: The item to find the predecessor of.

    Returns:
        The item before the given item in the sequence.

    Raises:
        ValueError: If the item is not found in the sequence or if the item


    Examples:
        >>> prev([1, 2, 3], 2)
        1
        >>> prev([1, 2, 3], 1)
        Traceback (most recent call last):
            ...
        ValueError: Item 1 has no predecessor.
        >>> prev([1, 2, 3], 4)
        Traceback (most recent call last):
            ...
        ValueError: Item not found in sequence: 4

    """
    try:
        index = sequence.index(item)
    except ValueError as e:
        msg = f"Item not found in sequence: {item!r}"
        raise ValueError(msg) from e

    if index == 0:
        msg = f"Item {item!r} has no predecessor."
        raise ValueError(msg)

    return sequence[index - 1]


def pairwise(
    iterable: Iterable[_T], /, default: _DT = None
) -> Generator[tuple[_T | _DT, _T]]:
    """Iterate over pairs of items in an iterable.

    Args:
        iterable: The iterable to iterate over.
        default: The default value to use for the first item.

    Yields:
        Pairs of items from the iterable. The first item in each pair is the
        previous item, or the default value if the first item is yielded.

    Examples:
        >>> list(pairwise([]))
        []
        >>> list(pairwise([1]))
        [(None, 1)]
        >>> list(pairwise([1, 2]))
        [(None, 1), (1, 2)]
        >>> list(pairwise([1, 2, 3]))
        [(None, 1), (1, 2), (2, 3)]

    """
    prev_item = default

    for item in iterable:
        yield prev_item, item
        prev_item = item


def is_first_index(sized: Sized, index: SupportsIndex) -> bool:
    """Check if an index resolves to the first index in a `Sized` object.

    Args:
        sized: The `Sized` object to check.
        index: The index to check.

    Returns:
        `True` if the index is the first index in the `Sized` object,
        `False` otherwise.

    Examples:
        >>> is_first_index([1, 2, 3], 0)
        True
        >>> is_first_index([1, 2, 3], 1)
        False
        >>> is_first_index([1, 2, 3], -1)
        False
        >>> is_first_index([1, 2, 3], -3)
        True
        >>> is_first_index([], 0)
        True

    """
    start, *_ = slice(index, index).indices(len(sized))

    return start == 0


def is_close(
    a: SupportsFloat | SupportsIndex, b: SupportsFloat | SupportsIndex, /
) -> bool:
    """Check if two floating-point numbers are almost equal.

    Args:
        a: The first number to compare.
        b: The second number to compare.

    Returns:
        `True` if the two numbers are almost equal, `False` otherwise.

    Examples:
        import math
        >>> is_close(1.0, 1.0)
        True
        >>> is_close(1.0, 1.0 + 1e-20)
        True
        >>> is_close(1.0, 1.0 + 0.001)
        False
        >>> is_close(0, math.sin(math.pi))
        True

    """
    return math.isclose(
        a,
        b,
        rel_tol=constants.FLOAT_RELATIVE_TOLERANCE,
        abs_tol=constants.FLOAT_ABSOLUTE_TOLERANCE,
    )


@overload
def apply_single_or_many(func: _Map[_T, _NT], value: _T, /) -> _NT: ...


@overload
def apply_single_or_many(
    func: _Map[_T, _NT], first: _T, second: _T, /, *values: _T
) -> tuple[_NT, ...]: ...


def apply_single_or_many(
    func: _Map[_T, _NT], /, *values: _T
) -> _NT | tuple[_NT, ...]:
    """Apply a function to one or more values.

    Args:
        func: The function to apply.
        values: The values to apply the function to.

    Returns:
        The result of applying the function to the value, or a tuple of
        such results if multiple values are provided.

    Examples:
        >>> apply_single_or_many(str, 1)
        '1'
        >>> apply_single_or_many(str, 1, 2, 3)
        ('1', '2', '3')

    """
    result = tuple(map(func, values))

    return result[0] if len(result) == 1 else result


def extract_function_name_and_args(attr: str) -> tuple[str, str] | None:
    """Extract function name and arguments from a function-call-like attribute.

    An attribute is considered to be a function call if it has the form
    `name(args)`. This function extracts the name and the arguments from such
    an attribute. If the attribute is not a function call, `None` is returned.

    Args:
    attr: The attribute to extract the function name and arguments from.

    Returns:
    A tuple containing the function name and the arguments,
    or `None` if the attribute is not a function call.

    Examples:
    >>> extract_function_name_and_args("foo()") is None  # no arguments
    True
    >>> extract_function_name_and_args("foo(42)")
    ('foo', '42')
    >>> extract_function_name_and_args("foo(42, 'bar')")
    ('foo', "42, 'bar'")
    >>> extract_function_name_and_args(
    ...     "bar"
    ... ) is None  # not a function call
    True

    """
    match = re.match(r"^([^\(\)]+)\(([^\(\)]+)\)$", attr)

    if match is None:
        return None

    return match.group(1), match.group(2)


def tan(degrees: float) -> float:
    """Compute the tangent of an angle in degrees.

    This function is a wrapper around `math.tan` that takes an angle in degrees
    and returns "nice" values for common angles.

    Args:
        degrees: The angle in degrees.

    Returns:
        The tangent of the angle.

    Examples:
        >>> tan(0)
        0
        >>> tan(45)
        1

    """
    degrees %= 180

    match degrees:
        case 0:
            return 0
        case 45:
            return 1
        case 135:
            return -1
        case _:
            return math.tan(math.radians(degrees))


def arctan(value: float) -> float:
    """Compute the arctangent of a value in degrees.

    This function is a wrapper around `math.atan` that returns "nice" values
    for common inputs. The output is in degrees.

    Args:
        value: The value to compute the arctangent of.

    Returns:
        The arctangent of the value in degrees.

    Examples:
        >>> arctan(0)
        0
        >>> arctan(1)
        45

    """
    match value:
        case 0:
            return 0
        case 1:
            return 45
        case -1:
            return -45
        case _:
            return math.degrees(math.atan(value))


def degrees(radians: float) -> float:
    """Convert radians to degrees, returning a value in the range (-180, 180].

    Args:
        radians: The angle in radians.

    Returns:
        The angle in degrees in the range (-180, 180].

    Examples:
        >>> from math import pi
        >>> degrees(0)
        0.0
        >>> degrees(pi)
        180.0
        >>> degrees(2 * pi)
        0.0
        >>> degrees(-pi / 2)
        -90.0
        >>> degrees(-pi)
        180.0

    """
    deg = math.degrees(radians)
    normalized = ((deg + 180) % 360) - 180

    if is_close(normalized, -180):
        return 180.0

    return normalized


def signum(x: float) -> Literal[-1, 0, 1]:
    """Compute the signum function `sgn(x)`.

    Args:
        x: The value to compute the signum of.

    Returns:
        -1 if the value is negative, 0 if the value is zero, and 1 if the value
        is positive.

    Examples:
        >>> signum(-5)
        -1
        >>> signum(0)
        0
        >>> signum(0.5)
        1

    """
    if x < 0:
        return -1

    if x > 0:
        return 1

    return 0


def arccos(value: float) -> float:
    """Compute the arccosine of a value in degrees.

    This function is a wrapper around `math.acos` that returns "nice" values
    for common inputs. The output is in degrees.

    Args:
        value: The value to compute the arccosine of.

    Returns:
        The arccosine of the value in degrees.

    Examples:
        >>> arccos(1)
        0
        >>> arccos(0)
        90

    """
    match value:
        case 1:
            return 0
        case 0:
            return 90
        case _:
            return math.degrees(math.acos(value))



from importlib import metadata as __metadata

import affine as __affine
from typing_extensions import Final as __Final

from svglab import constants as __constants
from svglab.attrparse.angle import Angle, AngleUnit
from svglab.attrparse.color import Color
from svglab.attrparse.d import (
    ArcTo,
    ClosePath,
    CubicBezierTo,
    D,
    HorizontalLineTo,
    LineTo,
    MoveTo,
    PathCommand,
    QuadraticBezierTo,
    SmoothCubicBezierTo,
    SmoothQuadraticBezierTo,
    VerticalLineTo,
)
from svglab.attrparse.length import Length
from svglab.attrparse.point import Point
from svglab.attrparse.points import Points
from svglab.attrparse.transform import (
    Matrix,
    Reifiable,
    Rotate,
    Scale,
    SkewX,
    SkewY,
    Transform,
    TransformFunction,
    Translate,
    compose,
)
from svglab.elements.common import (
    Element,
    Tag,
    TextElement,
    swap_transforms,
)
from svglab.elements.svg import Svg
from svglab.elements.tags import (
    A,
    AltGlyph,
    AltGlyphDef,
    AltGlyphItem,
    Animate,
    AnimateColor,
    AnimateMotion,
    AnimateTransform,
    Circle,
    ClipPath,
    ColorProfile,
    Cursor,
    Defs,
    Desc,
    Ellipse,
    FeBlend,
    FeColorMatrix,
    FeComponentTransfer,
    FeComposite,
    FeConvolveMatrix,
    FeDiffuseLighting,
    FeDisplacementMap,
    FeDistantLight,
    FeFlood,
    FeFuncA,
    FeFuncB,
    FeFuncG,
    FeFuncR,
    FeGaussianBlur,
    FeImage,
    FeMerge,
    FeMergeNode,
    FeMorphology,
    FeOffset,
    FePointLight,
    FeSpecularLighting,
    FeSpotLight,
    FeTile,
    FeTurbulence,
    Filter,
    Font,
    FontFace,
    FontFaceFormat,
    FontFaceName,
    FontFaceSrc,
    FontFaceUri,
    ForeignObject,
    G,
    Glyph,
    GlyphRef,
    Hkern,
    Image,
    Line,
    LinearGradient,
    Marker,
    Mask,
    Metadata,
    MissingGlyph,
    Mpath,
    Path,
    Pattern,
    Polygon,
    Polyline,
    RadialGradient,
    Rect,
    Script,
    Set,
    Stop,
    Style,
    Switch,
    Symbol,
    Text,
    TextPath,
    Title,
    Tref,
    Tspan,
    Use,
    View,
    Vkern,
)
from svglab.elements.text_elements import CData, Comment, RawText
from svglab.elements.traits import (
    BasicShape,
    ContainerElement,
    DescriptiveElement,
    FilterPrimitiveElement,
    GradientElement,
    GraphicsElement,
    GraphicsReferencingElement,
    LightSourceElement,
    Shape,
    StructuralElement,
    TextContentBlockElement,
    TextContentChildElement,
    TextContentElement,
)
from svglab.errors import (
    SvgElementNotFoundError,
    SvgError,
    SvgPathError,
    SvgPathMissingMoveToError,
    SvgReifyError,
    SvgTransformSwapError,
    SvgUnitConversionError,
)
from svglab.parse import parse_svg
from svglab.serialize import (
    DEFAULT_FORMATTER,
    MINIMAL_FORMATTER,
    FloatPrecisionSettings,
    Formatter,
    get_current_formatter,
    set_formatter,
)
from svglab.utiltypes import LengthUnit


__all__ = [
    "DEFAULT_FORMATTER",
    "MINIMAL_FORMATTER",
    "A",
    "AltGlyph",
    "AltGlyphDef",
    "AltGlyphItem",
    "Angle",
    "AngleUnit",
    "Animate",
    "AnimateColor",
    "AnimateMotion",
    "AnimateTransform",
    "ArcTo",
    "BasicShape",
    "CData",
    "Circle",
    "ClipPath",
    "ClosePath",
    "Color",
    "ColorProfile",
    "Comment",
    "ContainerElement",
    "CubicBezierTo",
    "Cursor",
    "D",
    "Defs",
    "Desc",
    "DescriptiveElement",
    "Element",
    "Ellipse",
    "FeBlend",
    "FeColorMatrix",
    "FeComponentTransfer",
    "FeComposite",
    "FeConvolveMatrix",
    "FeDiffuseLighting",
    "FeDisplacementMap",
    "FeDistantLight",
    "FeFlood",
    "FeFuncA",
    "FeFuncB",
    "FeFuncG",
    "FeFuncR",
    "FeGaussianBlur",
    "FeImage",
    "FeMerge",
    "FeMergeNode",
    "FeMorphology",
    "FeOffset",
    "FePointLight",
    "FeSpecularLighting",
    "FeSpotLight",
    "FeTile",
    "FeTurbulence",
    "Filter",
    "FilterPrimitiveElement",
    "FloatPrecisionSettings",
    "Font",
    "FontFace",
    "FontFaceFormat",
    "FontFaceName",
    "FontFaceSrc",
    "FontFaceUri",
    "ForeignObject",
    "Formatter",
    "G",
    "Glyph",
    "GlyphRef",
    "GradientElement",
    "GraphicsElement",
    "GraphicsReferencingElement",
    "Hkern",
    "HorizontalLineTo",
    "Image",
    "Length",
    "LengthUnit",
    "LightSourceElement",
    "Line",
    "LineTo",
    "LinearGradient",
    "Marker",
    "Mask",
    "Matrix",
    "Metadata",
    "MissingGlyph",
    "MoveTo",
    "Mpath",
    "Path",
    "PathCommand",
    "Pattern",
    "Point",
    "Points",
    "Polygon",
    "Polyline",
    "QuadraticBezierTo",
    "RadialGradient",
    "RawText",
    "Rect",
    "Reifiable",
    "Rotate",
    "Scale",
    "Script",
    "Set",
    "Shape",
    "SkewX",
    "SkewY",
    "SmoothCubicBezierTo",
    "SmoothQuadraticBezierTo",
    "Stop",
    "StructuralElement",
    "Style",
    "Svg",
    "SvgElementNotFoundError",
    "SvgError",
    "SvgPathError",
    "SvgPathMissingMoveToError",
    "SvgReifyError",
    "SvgTransformSwapError",
    "SvgUnitConversionError",
    "Switch",
    "Symbol",
    "Tag",
    "Text",
    "TextContentBlockElement",
    "TextContentChildElement",
    "TextContentElement",
    "TextElement",
    "TextPath",
    "Title",
    "Transform",
    "TransformFunction",
    "Translate",
    "Tref",
    "Tspan",
    "Use",
    "VerticalLineTo",
    "View",
    "Vkern",
    "compose",
    "get_current_formatter",
    "parse_svg",
    "set_formatter",
    "swap_transforms",
]


__affine.set_epsilon(__constants.FLOAT_ABSOLUTE_TOLERANCE)

__version__: __Final = __metadata.version(__name__)



from __future__ import annotations

import itertools

from typing_extensions import final, override

from svglab import models
from svglab.attrparse import d, length, point
from svglab.attrs import groups, regular
from svglab.elements import traits


# ! WARNING: `Element` and `ContainerElement` must always go last


def _length_or_zero(value: length.Length | None, /) -> length.Length:
    return value if value is not None else length.Length(0)


@final
class Path(
    groups.ConditionalProcessing,
    regular.Class,
    regular.D,
    regular.ExternalResourcesRequired,
    regular.Style,
    traits.Shape,
    traits.Element,
):
    pass


def _basic_shape_to_path(basic_shape: traits.BasicShape, /) -> Path:
    """Convert a basic shape to a `Path` element."""
    # try to convert to D first, so we don't call convert() if the shape
    # is not convertible
    d = basic_shape.to_d()

    path = models.convert(basic_shape, Path)
    path.d = d

    return path


def _points_to_d(element: regular.Points) -> d.D:
    """Convert an element with a `points` attribute to a `D` instance."""
    points = element.points if element.points is not None else []
    path_data = d.D()

    if points:
        path_data.move_to(points[0])

        for point in itertools.islice(points, 1, None):
            path_data.line_to(point)

    return path_data


def _ellipse_to_d(
    *,
    cx: length.Length,
    cy: length.Length,
    rx: length.Length,
    ry: length.Length,
) -> d.D:
    """Convert an ellipse to a `D` instance.

    Args:
        cx: The x-coordinate of the center of the ellipse.
        cy: The y-coordinate of the center of the ellipse.
        rx: The x-radius of the ellipse.
        ry: The y-radius of the ellipse.

    Returns:
        A `D` instance representing the ellipse.

    References:
        https://stackoverflow.com/a/10477334
        https://aleen42.gitbooks.io/wiki/content/Programming/JavaScript/webgl/SVG/convert_shapes_to_path/convert_shapes_to_path.html

    """
    return (
        d.D()
        .move_to(point.Point(cx - rx, cy))
        .arc_to(
            point.Point(rx, ry),
            0,
            point.Point(2 * rx, 0),
            large=True,
            sweep=False,
            relative=True,
        )
        .arc_to(
            point.Point(rx, ry),
            0,
            point.Point(-2 * rx, 0),
            large=True,
            sweep=False,
            relative=True,
        )
    )


@final
class A(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Style,
    regular.Target,
    traits.ContainerElement,
):
    pass


@final
class AltGlyph(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.Class,
    regular.DxListOfLengths,
    regular.DyListOfLengths,
    regular.ExternalResourcesRequired,
    regular.Format,
    regular.GlyphRef,
    regular.RotateListOfNumbers,
    regular.Style,
    regular.XListOfCoordinates,
    regular.YListOfCoordinates,
    traits.TextContentChildElement,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class AltGlyphDef(traits.Element):
    pass


@final
class AltGlyphItem(traits.Element):
    pass


@final
class Animate(
    groups.AnimationAddition,
    groups.AnimationAttributeTarget,
    groups.AnimationValue,
    groups.ConditionalProcessing,
    groups.Xlink,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class AnimateColor(
    groups.AnimationAddition,
    groups.AnimationAttributeTarget,
    groups.AnimationValue,
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.ExternalResourcesRequired,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class AnimateMotion(
    groups.AnimationAddition,
    groups.AnimationValue,
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.ExternalResourcesRequired,
    regular.KeyPoints,
    regular.Origin,
    regular.Path,
    regular.RotateNumberAutoAutoReverse,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class AnimateTransform(
    groups.AnimationAddition,
    groups.AnimationAttributeTarget,
    groups.AnimationValue,
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.ExternalResourcesRequired,
    regular.TypeAnimateTransform,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class Circle(
    groups.ConditionalProcessing,
    regular.Class,
    regular.Cx,
    regular.Cy,
    regular.ExternalResourcesRequired,
    regular.R,
    regular.Style,
    traits.BasicShape,
    traits.Element,
):
    @override
    def to_d(self) -> d.D:
        cx = _length_or_zero(self.cx)
        cy = _length_or_zero(self.cy)
        r = _length_or_zero(self.r)

        return _ellipse_to_d(cx=cx, cy=cy, rx=r, ry=r)

    @override
    def to_path(self) -> Path:
        return _basic_shape_to_path(self)


@final
class ClipPath(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ClipPathUnits,
    regular.ExternalResourcesRequired,
    regular.Style,
    traits.Element,
):
    pass


@final
class ColorProfile(
    groups.Xlink,
    regular.Local,
    regular.NameName,
    regular.RenderingIntent,
    traits.Element,
):
    pass


@final
class Cursor(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.ExternalResourcesRequired,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.Element,
):
    pass


@final
class Defs(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Style,
    traits.StructuralElement,
    traits.ContainerElement,
):
    pass


@final
class Desc(
    regular.Class, regular.Style, traits.DescriptiveElement, traits.Element
):
    pass


@final
class Ellipse(
    groups.ConditionalProcessing,
    regular.Class,
    regular.Cx,
    regular.Cy,
    regular.ExternalResourcesRequired,
    regular.Rx,
    regular.Ry,
    regular.Style,
    traits.BasicShape,
    traits.Element,
):
    @override
    def to_d(self) -> d.D:
        cx = _length_or_zero(self.cx)
        cy = _length_or_zero(self.cy)
        rx = _length_or_zero(self.rx)
        ry = _length_or_zero(self.ry)

        return _ellipse_to_d(cx=cx, cy=cy, rx=rx, ry=ry)

    @override
    def to_path(self) -> Path:
        return _basic_shape_to_path(self)


@final
class FeBlend(
    regular.Class,
    regular.In,
    regular.In2,
    regular.Mode,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeColorMatrix(
    regular.Class,
    regular.In,
    regular.In2,
    regular.Style,
    regular.TypeFeColorMatrix,
    regular.ValuesListOfNumbers,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeComponentTransfer(
    regular.Class,
    regular.In,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeComposite(
    regular.Class,
    regular.In,
    regular.In2,
    regular.K1,
    regular.K2,
    regular.K3,
    regular.K4,
    regular.OperatorFeComposite,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeConvolveMatrix(
    regular.Bias,
    regular.Class,
    regular.Divisor,
    regular.In,
    regular.KernelMatrix,
    regular.KernelUnitLength,
    regular.Order,
    regular.PreserveAlpha,
    regular.Style,
    regular.TargetX,
    regular.TargetY,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeDiffuseLighting(
    regular.Class,
    regular.DiffuseConstant,
    regular.In,
    regular.KernelUnitLength,
    regular.Style,
    regular.SurfaceScale,
    traits.Element,
):
    pass


@final
class FeDisplacementMap(
    regular.Class,
    regular.In,
    regular.In2,
    regular.Scale,
    regular.Style,
    regular.XChannelSelector,
    regular.YChannelSelector,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeDistantLight(
    regular.Azimuth,
    regular.Elevation,
    traits.LightSourceElement,
    traits.Element,
):
    pass


@final
class FeFlood(
    regular.Class,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeFuncA(groups.TransferFunction, traits.Element):
    pass


@final
class FeFuncB(groups.TransferFunction, traits.Element):
    pass


@final
class FeFuncG(groups.TransferFunction, traits.Element):
    pass


@final
class FeFuncR(groups.TransferFunction, traits.Element):
    pass


@final
class FeGaussianBlur(
    regular.Class,
    regular.In,
    regular.StdDeviation,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeImage(
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.PreserveAspectRatio,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeMerge(
    regular.Class,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeMergeNode(regular.In, traits.Element):
    pass


@final
class FeMorphology(
    regular.Class,
    regular.In,
    regular.OperatorFeMorphology,
    regular.Radius,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeOffset(
    regular.Class,
    regular.DxNumber,
    regular.DyNumber,
    regular.In,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FePointLight(
    regular.XNumber,
    regular.YNumber,
    regular.Z,
    traits.LightSourceElement,
    traits.Element,
):
    pass


@final
class FeSpecularLighting(
    regular.Class,
    regular.In,
    regular.KernelUnitLength,
    regular.SpecularConstant,
    regular.SpecularExponent,
    regular.Style,
    regular.SurfaceScale,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeSpotLight(
    regular.LimitingConeAngle,
    regular.PointsAtX,
    regular.PointsAtY,
    regular.PointsAtZ,
    regular.SpecularExponent,
    regular.XNumber,
    regular.YNumber,
    regular.Z,
    traits.LightSourceElement,
    traits.Element,
):
    pass


@final
class FeTile(
    regular.Class,
    regular.In,
    regular.Style,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeTurbulence(
    regular.BaseFrequency,
    regular.Class,
    regular.NumOctaves,
    regular.Seed,
    regular.StitchTiles,
    regular.Style,
    regular.TypeFeTurbluence,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class Filter(
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.FilterRes,
    regular.FilterUnits,
    regular.Height,
    regular.PrimitiveUnits,
    regular.Style,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.Element,
):
    pass


@final
class Font(
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.HorizAdvX,
    regular.HorizOriginX,
    regular.HorizOriginY,
    regular.Style,
    regular.VertAdvY,
    regular.VertOriginX,
    regular.VertOriginY,
    traits.Element,
):
    pass


@final
class FontFace(
    regular.AccentHeight,
    regular.Alphabetic,
    regular.Ascent,
    regular.Bbox,
    regular.CapHeight,
    regular.Descent,
    regular.Hanging,
    regular.Ideographic,
    regular.Mathematical,
    regular.OverlinePosition,
    regular.OverlineThickness,
    regular.Panose1,
    regular.Slope,
    regular.Stemh,
    regular.Stemv,
    regular.StrikethroughPosition,
    regular.StrikethroughThickness,
    regular.UnderlinePosition,
    regular.UnderlineThickness,
    regular.UnicodeRange,
    regular.UnitsPerEm,
    regular.VAlphabetic,
    regular.VHanging,
    regular.VIdeographic,
    regular.VMathematical,
    regular.Widths,
    regular.XHeight,
    traits.Element,
):
    pass


@final
class FontFaceFormat(regular.String, traits.Element):
    pass


@final
class FontFaceName(regular.String, traits.Element):
    pass


@final
class FontFaceSrc(traits.Element):
    pass


@final
class FontFaceUri(groups.Xlink, traits.Element):
    pass


@final
class ForeignObject(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.Style,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.Element,
):
    pass


@final
class G(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Style,
    traits.StructuralElement,
    traits.ContainerElement,
):
    pass


@final
class Glyph(
    regular.ArabicForm,
    regular.Class,
    regular.D,
    regular.GlyphName,
    regular.HorizAdvX,
    regular.LangGlyph,
    regular.Orientation,
    regular.Style,
    regular.Unicode,
    regular.VertAdvY,
    regular.VertOriginX,
    regular.VertOriginY,
    traits.ContainerElement,
):
    pass


@final
class GlyphRef(
    groups.Xlink,
    regular.Class,
    regular.DxNumber,
    regular.DyNumber,
    regular.Format,
    regular.GlyphRef,
    regular.Style,
    regular.XNumber,
    regular.YNumber,
    traits.Element,
):
    pass


@final
class Hkern(
    regular.G1,
    regular.G2,
    regular.K,
    regular.U1,
    regular.U2,
    traits.Element,
):
    pass


@final
class Image(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.PreserveAspectRatio,
    regular.Style,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.GraphicsElement,
    traits.GraphicsReferencingElement,
    traits.Element,
):
    pass


@final
class Line(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Style,
    regular.X1,
    regular.X2,
    regular.Y1,
    regular.Y2,
    traits.BasicShape,
    traits.Element,
):
    @override
    def to_d(self) -> d.D:
        x1 = _length_or_zero(self.x1)
        y1 = _length_or_zero(self.y1)
        x2 = _length_or_zero(self.x2)
        y2 = _length_or_zero(self.y2)

        return (
            d.D().move_to(point.Point(x1, y1)).line_to(point.Point(x2, y2))
        )

    @override
    def to_path(self) -> Path:
        return _basic_shape_to_path(self)


@final
class LinearGradient(
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.GradientTransform,
    regular.GradientUnits,
    regular.SpreadMethod,
    regular.Style,
    regular.X1,
    regular.X2,
    regular.Y1,
    regular.Y2,
    traits.GradientElement,
    traits.Element,
):
    pass


@final
class Marker(
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.MarkerHeight,
    regular.MarkerUnits,
    regular.MarkerWidth,
    regular.Orient,
    regular.PreserveAspectRatio,
    regular.RefX,
    regular.RefY,
    regular.Style,
    regular.ViewBox,
    traits.ContainerElement,
):
    pass


@final
class Mask(
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.MaskContentUnits,
    regular.MaskUnits,
    regular.Style,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.ContainerElement,
):
    pass


@final
class Metadata(traits.DescriptiveElement, traits.Element):
    pass


@final
class MissingGlyph(
    regular.Class,
    regular.D,
    regular.HorizAdvX,
    regular.Style,
    regular.VertAdvY,
    regular.VertOriginX,
    regular.VertOriginY,
    traits.ContainerElement,
):
    pass


@final
class Mpath(
    groups.Xlink, regular.ExternalResourcesRequired, traits.Element
):
    pass


@final
class Pattern(
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.PatternContentUnits,
    regular.PatternTransform,
    regular.PatternUnits,
    regular.PreserveAspectRatio,
    regular.Style,
    regular.ViewBox,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.ContainerElement,
):
    pass


@final
class Polygon(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Points,
    regular.Style,
    traits.BasicShape,
    traits.Element,
):
    @override
    def to_d(self) -> d.D:
        return _points_to_d(self).close()

    @override
    def to_path(self) -> Path:
        return _basic_shape_to_path(self)


@final
class Polyline(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Points,
    regular.Style,
    traits.BasicShape,
    traits.Element,
):
    @override
    def to_d(self) -> d.D:
        return _points_to_d(self)

    @override
    def to_path(self) -> Path:
        return _basic_shape_to_path(self)


@final
class RadialGradient(
    groups.Xlink,
    regular.Class,
    regular.Cx,
    regular.Cy,
    regular.ExternalResourcesRequired,
    regular.Fr,
    regular.Fx,
    regular.Fy,
    regular.GradientTransform,
    regular.GradientUnits,
    regular.R,
    regular.SpreadMethod,
    regular.Style,
    traits.GradientElement,
    traits.Element,
):
    pass


@final
class Rect(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.Rx,
    regular.Ry,
    regular.Style,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.BasicShape,
    traits.Element,
):
    @override
    def to_d(self) -> d.D:
        x = _length_or_zero(self.x)
        y = _length_or_zero(self.y)
        width = _length_or_zero(self.width)
        height = _length_or_zero(self.height)

        match self.rx, self.ry:
            case None, None:
                rx = ry = length.Length(0)
            case length.Length(), None:
                rx = ry = self.rx
            case None, length.Length():
                rx = ry = self.ry
            case length.Length(), length.Length():
                rx = self.rx
                ry = self.ry

        return (
            d.D()
            .move_to(point.Point(x + rx, y))
            .horizontal_line_to(x + width - rx)
            .arc_to(
                point.Point(rx, ry),
                0,
                point.Point(x + width, y + ry),
                large=False,
                sweep=True,
            )
            .vertical_line_to(y + height - ry)
            .arc_to(
                point.Point(rx, ry),
                0,
                point.Point(x + width - rx, y + height),
                large=False,
                sweep=True,
            )
            .horizontal_line_to(x + rx)
            .arc_to(
                point.Point(rx, ry),
                0,
                point.Point(x, y + height - ry),
                large=False,
                sweep=True,
            )
            .vertical_line_to(y + ry)
            .arc_to(
                point.Point(rx, ry),
                0,
                point.Point(x + rx, y),
                large=False,
                sweep=True,
            )
        )

    @override
    def to_path(self) -> Path:
        return _basic_shape_to_path(self)


@final
class Script(
    groups.Xlink,
    regular.ExternalResourcesRequired,
    regular.TypeContentType,
    traits.Element,
):
    pass


@final
class Set(
    groups.AnimationAttributeTarget,
    groups.Xlink,
    regular.ExternalResourcesRequired,
    regular.To,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class Stop(
    regular.Class,
    regular.OffsetNumberPercentage,
    regular.Style,
    traits.Element,
):
    pass


@final
class Style(
    regular.Media, regular.Title, regular.TypeContentType, traits.Element
):
    pass


@final
class Switch(
    groups.ConditionalProcessing,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Style,
    traits.ContainerElement,
):
    pass


@final
class Symbol(
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.PreserveAspectRatio,
    regular.RefX,
    regular.RefY,
    regular.Style,
    regular.ViewBox,
    traits.StructuralElement,
    traits.ContainerElement,
):
    pass


@final
class Text(
    groups.ConditionalProcessing,
    regular.Class,
    regular.DxListOfLengths,
    regular.DyListOfLengths,
    regular.ExternalResourcesRequired,
    regular.LengthAdjust,
    regular.RotateListOfNumbers,
    regular.Style,
    regular.TextLength,
    regular.XListOfCoordinates,
    regular.YListOfCoordinates,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class TextPath(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Method,
    regular.Spacing,
    regular.StartOffset,
    regular.Style,
    traits.TextContentChildElement,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class Title(
    regular.Class, regular.Style, traits.DescriptiveElement, traits.Element
):
    pass


@final
class Tref(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Style,
    traits.TextContentChildElement,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class Tspan(
    groups.ConditionalProcessing,
    regular.Class,
    regular.DxListOfLengths,
    regular.DyListOfLengths,
    regular.ExternalResourcesRequired,
    regular.LengthAdjust,
    regular.RotateListOfNumbers,
    regular.Style,
    regular.TextLength,
    regular.XListOfCoordinates,
    regular.YListOfCoordinates,
    traits.TextContentBlockElement,
    traits.TextContentChildElement,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class Use(
    groups.ConditionalProcessing,
    groups.Xlink,
    regular.Class,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.Style,
    regular.Width,
    regular.XCoordinate,
    regular.YCoordinate,
    traits.GraphicsElement,
    traits.GraphicsReferencingElement,
    traits.StructuralElement,
    traits.Element,
):
    pass


@final
class View(
    regular.ExternalResourcesRequired,
    regular.PreserveAspectRatio,
    regular.ViewBox,
    regular.ViewTarget,
    regular.ZoomAndPan,
    traits.Element,
):
    pass


@final
class Vkern(
    regular.G1,
    regular.G2,
    regular.K,
    regular.U1,
    regular.U2,
    traits.Element,
):
    pass



from __future__ import annotations

import abc
import collections
import reprlib
import sys
from collections.abc import Generator, Mapping

import bs4
import pydantic
from typing_extensions import (
    Final,
    Self,
    SupportsIndex,
    TypeVar,
    cast,
    overload,
    override,
)

from svglab import constants, errors, models, serialize, utils
from svglab.attrparse import length, transform
from svglab.attrs import common as common_attrs
from svglab.attrs import groups, presentation, regular
from svglab.attrs import names as attr_names
from svglab.elements import names


_T = TypeVar("_T")
_T_tag = TypeVar("_T_tag", bound="Tag")
_TransformT1 = TypeVar("_TransformT1", bound=transform.TransformFunction)
_TransformT2 = TypeVar("_TransformT2", bound=transform.TransformFunction)

_EMPTY_PARAM: Final = object()
"""A sentinel value for an empty parameter."""


class StrokeWidthScaled:
    """The element's `stroke-width` attribute should be scaled."""


def _scale_attr(attr: _T, /, by: float) -> _T:
    """Scale an attribute by the given factor.

    This is a helper function for scaling attributes. If the attribute is a
    number or a non-percentage length, it is scaled by the given factor.
    Otherwise, the attribute is left unchanged. If the attribute is a list or
    tuple, the function is applied recursively to each element.

    Args:
    attr: The attribute to scale.
    by: The factor by which to scale the attribute.

    Returns:
    The scaled attribute.

    Examples:
    >>> from svglab import Length
    >>> _scale_attr(Length(10), 2)
    Length(value=20.0, unit=None)
    >>> _scale_attr(None, 2) is None
    True
    >>> _scale_attr(Length(10, "%"), 2)
    Length(value=10.0, unit='%')

    """
    match attr:
        case length.Length(_, "%"):
            return attr
        case int() | float() | length.Length():
            return cast(_T, attr * by)
        case list() | tuple():
            return type(attr)(_scale_attr(item, by) for item in attr)
        case _:
            return attr


def _scale_stroke_width(tag: presentation.StrokeWidth, by: float) -> None:
    if (
        isinstance(tag, presentation.VectorEffect)
        and tag.vector_effect == "non-scaling-stroke"
    ):
        return

    sw_set = tag.stroke_width is not None

    if not sw_set and not isinstance(tag, StrokeWidthScaled):
        return

    if not sw_set:
        tag.stroke_width = length.Length(1)

    tag.stroke_width = _scale_attr(tag.stroke_width, by)  # type: ignore[reportAttributeAccessIssue]

    # if stroke-width was not set and the scaled value is 1 (default),
    # remove the attribute
    if (
        not sw_set
        and isinstance(tag.stroke_width, length.Length)
        and utils.is_close(float(tag.stroke_width), 1)
    ):
        tag.stroke_width = None


def scale_distance_along_a_path_attrs(tag: object, by: float) -> None:
    if isinstance(tag, presentation.StrokeDasharray) and isinstance(
        tag.stroke_dasharray, list
    ):
        tag.stroke_dasharray = [
            _scale_attr(dash, by) for dash in tag.stroke_dasharray
        ]
    if isinstance(tag, presentation.StrokeDashoffset):
        tag.stroke_dashoffset = _scale_attr(tag.stroke_dashoffset, by)


def _scale(tag: object, scale: transform.Scale) -> None:  # noqa: PLR0915
    if not utils.is_close(scale.sx, scale.sy):
        raise ValueError("Non-uniform scaling is not supported.")

    factor = scale.sx

    if utils.is_close(factor, 1):
        return

    if isinstance(tag, regular.Width):
        tag.width = _scale_attr(tag.width, factor)
    if isinstance(tag, regular.Height):
        tag.height = _scale_attr(tag.height, factor)
    if isinstance(tag, regular.R):
        tag.r = _scale_attr(tag.r, factor)
    if isinstance(tag, regular.X1):
        tag.x1 = _scale_attr(tag.x1, factor)
    if isinstance(tag, regular.Y1):
        tag.y1 = _scale_attr(tag.y1, factor)
    if isinstance(tag, regular.X2):
        tag.x2 = _scale_attr(tag.x2, factor)
    if isinstance(tag, regular.Y2):
        tag.y2 = _scale_attr(tag.y2, factor)
    if isinstance(tag, regular.Rx):
        tag.rx = _scale_attr(tag.rx, factor)
    if isinstance(tag, regular.Ry):
        tag.ry = _scale_attr(tag.ry, factor)
    if isinstance(tag, regular.Cx):
        tag.cx = _scale_attr(tag.cx, factor)
    if isinstance(tag, regular.Cy):
        tag.cy = _scale_attr(tag.cy, factor)
    if isinstance(tag, regular.Fx):
        tag.fx = _scale_attr(tag.fx, factor)
    if isinstance(tag, regular.Fy):
        tag.fy = _scale_attr(tag.fy, factor)
    if isinstance(tag, common_attrs.FontSize):
        tag.font_size = _scale_attr(tag.font_size, factor)
    if isinstance(tag, regular.Points) and tag.points is not None:
        tag.points = [scale @ point for point in tag.points]
    if isinstance(tag, regular.D) and tag.d is not None:
        tag.d = scale @ tag.d

    # these assignments have to be mutually exclusive, because the
    # type checker doesn't know that x being a <number> implies that x is
    # not a <coordinate> and vice versa
    if isinstance(tag, regular.XNumber):  # noqa: SIM114
        tag.x = _scale_attr(tag.x, factor)
    elif isinstance(tag, regular.XCoordinate):  # noqa: SIM114
        tag.x = _scale_attr(tag.x, factor)
    elif isinstance(tag, regular.XListOfCoordinates):
        tag.x = _scale_attr(tag.x, factor)

    if isinstance(tag, regular.YNumber):  # noqa: SIM114
        tag.y = _scale_attr(tag.y, factor)
    elif isinstance(tag, regular.YCoordinate):  # noqa: SIM114
        tag.y = _scale_attr(tag.y, factor)
    elif isinstance(tag, regular.YListOfCoordinates):
        tag.y = _scale_attr(tag.y, factor)

    if isinstance(tag, presentation.StrokeWidth):
        _scale_stroke_width(tag, factor)

    # no need to scale distance-along-a-path attributes if a custom path
    # length is provided because those attributes and pathLength are
    # proportional
    if not isinstance(tag, regular.PathLength):
        scale_distance_along_a_path_attrs(tag, factor)


def _translate_attr(attr: _T, /, by: float) -> _T:
    """Translate an attribute by the given amount.

    This is a helper function for translating attributes. If the attribute is
    a number or a non-percentage length, it is translated by the given amount.
    Otherwise, the attribute is left unchanged. If the attribute is a list or
    tuple, the function is applied recursively to each element.

    Args:
    attr: The attribute to translate.
    by: The amount by which to translate the attribute.

    Returns:
    The translated attribute.

    Examples:
    >>> from svglab import Length
    >>> _translate_attr(Length(10), 5)
    Length(value=15.0, unit=None)
    >>> _translate_attr(None, 5) is None
    True
    >>> _translate_attr(Length(10, "%"), 5)
    Length(value=10.0, unit='%')

    """
    match attr:
        case length.Length(_, "%"):
            return attr
        case int() | float():
            return cast(_T, attr + by)
        case length.Length():
            return attr + length.Length(by)
        case list() | tuple():
            return type(attr)(_translate_attr(item, by) for item in attr)
        case _:
            return attr


# pyright goes nuts if `tag` is annotated as `Element`... seems like a bug
# luckily we don't really care
def _translate(tag: object, translate: transform.Translate) -> None:
    tx, ty = translate.tx, translate.ty

    if utils.is_close(tx, 0) and utils.is_close(ty, 0):
        return

    zero = length.Length.zero()

    # these attributes are mandatory for the respective elements
    if isinstance(tag, regular.X1):
        tag.x1 = _translate_attr(tag.x1, tx)
    if isinstance(tag, regular.Y1):
        tag.y1 = _translate_attr(tag.y1, ty)
    if isinstance(tag, regular.X2):
        tag.x2 = _translate_attr(tag.x2, tx)
    if isinstance(tag, regular.Y2):
        tag.y2 = _translate_attr(tag.y2, ty)

    # but these are not, so if they are not present, we initialize them
    # to 0
    if isinstance(tag, regular.Cx):
        tag.cx = _translate_attr(tag.cx or zero, tx)
    if isinstance(tag, regular.Cy):
        tag.cy = _translate_attr(tag.cy or zero, ty)

    if isinstance(tag, regular.XNumber):
        tag.x = _translate_attr(tag.x or 0, tx)
    elif isinstance(tag, regular.XCoordinate):
        tag.x = _translate_attr(tag.x or zero, tx)
    elif isinstance(tag, regular.XListOfCoordinates):
        tag.x = _translate_attr(tag.x, tx)

    if isinstance(tag, regular.YNumber):
        tag.y = _translate_attr(tag.y or 0, ty)
    elif isinstance(tag, regular.YCoordinate):
        tag.y = _translate_attr(tag.y or zero, ty)
    elif isinstance(tag, regular.YListOfCoordinates):
        tag.y = _translate_attr(tag.y, ty)

    if isinstance(tag, regular.Points) and tag.points is not None:
        tag.points = [translate @ point for point in tag.points]

    if isinstance(tag, regular.D) and tag.d is not None:
        tag.d = translate @ tag.d


def swap_transforms(
    a: _TransformT1, b: _TransformT2, /
) -> tuple[_TransformT2, _TransformT1]:
    """Swap transforms, adjusting parameters so that the result is equal.

    Args:
        a: The first transform.
        b: The second transform.

    Returns:
        A 2-tuple (b', a') where b' and a' are the adjusted transforms.

    Raises:
        SvgTransformSwapError: If the transforms cannot be swapped.

    Examples:
        >>> from svglab.attrparse.transform import Scale, Translate, SkewX
        >>> swap_transforms(Translate(10, 20), Scale(2, 3))
        (Scale(sx=2.0, sy=3.0), Translate(tx=5.0, ty=6.666666666666667))
        >>> swap_transforms(Scale(2, 3), SkewX(45))
        (SkewX(angle=33.690067525979785), Scale(sx=2.0, sy=3.0))
        >>> swap_transforms(SkewX(45), Translate(10, 20))
        (Translate(tx=30.0, ty=20.0), SkewX(angle=45.0))

    """
    match a, b:
        # transformations of the same type
        case (transform.Translate(), transform.Translate()) | (
            transform.Scale(),
            transform.Scale(),
        ):
            return b, a

        # translate <-> scale
        case transform.Translate(tx, ty), transform.Scale(sx, sy) as scale:
            return scale, type(a)(tx / sx, ty / sy)

        case transform.Scale(sx, sy) as scale, transform.Translate(tx, ty):
            return type(b)(sx * tx, sy * ty), scale

        # translate <-> rotate
        case transform.Rotate(angle, cx, cy), transform.Translate(tx, ty):
            return type(b)(tx, ty), type(a)(angle, cx - tx, cy - ty)
        case transform.Translate(tx, ty), transform.Rotate(angle, cx, cy):
            return type(b)(angle, cx + tx, cy + ty), type(a)(tx, ty)

        # scale <-> rotate
        case transform.Rotate(angle, cx, cy), transform.Scale(
            sx, sy
        ) as scale:
            return scale, type(a)(angle, cx / sx, cy / sy)
        case transform.Scale(sx, sy) as scale, transform.Rotate(
            angle, cx, cy
        ):
            return type(b)(angle, cx * sx, cy * sy), scale

        # translate <-> skew
        case transform.SkewX(angle) as skew_x, transform.Translate(tx, ty):
            return type(b)(tx + ty * utils.tan(angle), ty), skew_x

        case transform.Translate(tx, ty), transform.SkewX(angle) as skew_x:
            return skew_x, type(a)(tx - ty * utils.tan(angle), ty)

        case transform.SkewY(angle) as skew_y, transform.Translate(tx, ty):
            return type(b)(tx, ty + tx * utils.tan(angle)), skew_y

        case transform.Translate(tx, ty), transform.SkewY(angle) as skew_y:
            return skew_y, type(a)(tx, ty - tx * utils.tan(angle))

        # scale <-> skew
        case transform.Scale(sx, sy) as scale, transform.SkewX(angle):
            if utils.is_close(sx, sy):
                return b, a

            angle = utils.arctan(sx / sy * utils.tan(angle))
            return type(b)(angle), scale

        case transform.SkewX(angle), transform.Scale(sx, sy) as scale:
            if utils.is_close(sx, sy):
                return b, a

            angle = utils.arctan(sy / sx * utils.tan(angle))
            return scale, type(a)(angle)

        case transform.Scale(sx, sy) as scale, transform.SkewY(angle):
            if utils.is_close(sx, sy):
                return b, a

            angle = utils.arctan(sy / sx * utils.tan(angle))
            return type(b)(angle), scale

        case transform.SkewY(angle), transform.Scale(sx, sy) as scale:
            if utils.is_close(sx, sy):
                return b, a

            angle = utils.arctan(sx / sy * utils.tan(angle))

            return scale, type(a)(angle)
        case _:
            raise errors.SvgTransformSwapError(a, b)


def _move_transformation_to_end(
    transform: transform.Transform, index: int
) -> None:
    """Move a transformation to the end of the transform list.

    This function moves a transformation from the given index to the end of
    the list, swapping it with each transformation that follows it.
    The transformations in the list are adjusted so that the result is the
    same. The transformation itself may have its parameters adjusted as well.

    Args:
        transform: A list of transformations.
        index: The index of the transformation to move.

    Raises:
        ValueError: If the index is out of range.
        SvgTransformSwapError: If two transformations cannot be swapped.

    Examples:
        >>> from svglab.attrparse.transform import Translate, Scale
        >>> transform = [Translate(10, 20), Scale(2, 3)]
        >>> _move_transformation_to_end(transform, 0)
        >>> transform
        [Scale(sx=2.0, sy=3.0), Translate(tx=5.0, ty=6.666666666666667)]

    """
    if not (0 <= index < len(transform)):
        msg = f"Index {index=} out of range"
        raise ValueError(msg)

    for i in range(index, len(transform) - 1):
        transform[i], transform[i + 1] = swap_transforms(
            transform[i], transform[i + 1]
        )


def _apply_transformation(
    tag: Tag, transformation: transform.TransformFunction, /
) -> None:
    """Apply a transformation to the attributes of the tag.

    Args:
    tag: The tag to transform.
    transformation: The transformation to apply.

    Raises:
    ValueError: If the transformation is not supported.
    SvgLengthConversionError: If a length attribute is not convertible
    to user units.

    """
    match transformation:
        case transform.Translate():
            _translate(tag, transformation)
        case transform.Scale():
            _scale(tag, transformation)
        case _:
            msg = f"Unsupported transformation: {transformation}"
            raise ValueError(msg)


def tag_name(tag: Tag | type[Tag], /) -> names.TagName:
    """Get the SVG tag name of the given tag or tag class.

    Args:
    tag: The tag or tag class.

    Returns:
    The SVG tag name.

    Examples:
    >>> from svglab import Rect
    >>> tag_name(Rect)
    'rect'

    """
    tag_cls = tag if isinstance(tag, type) else type(tag)

    return names.TAG_NAME_TO_NORMALIZED.inverse[tag_cls.__name__]


class Element(models.BaseModel, metaclass=abc.ABCMeta):
    """The base class of the SVG element hierarchy."""

    parent: Tag | None = pydantic.Field(default=None, init=False)

    def to_xml(
        self,
        *,
        pretty: bool = True,
        formatter: serialize.Formatter | None = None,
    ) -> str:
        """Convert the element to XML.

        Args:
        pretty: Whether to produce pretty-printed XML.
        indent: The number of spaces to indent each level of the document.
        formatter: The formatter to use for serialization.

        Returns:
        The XML representation of the element.

        Examples:
        >>> from svglab import Rect, Length
        >>> rect = Rect(id="foo", stroke_linecap="round")
        >>> rect.to_xml()
        '<rect id="foo" stroke-linecap="round"/>'

        """
        with formatter or serialize.get_current_formatter():
            soup = self.to_beautifulsoup_object()
            return utils.beautifulsoup_to_str(soup, pretty=pretty)

    @abc.abstractmethod
    def to_beautifulsoup_object(self) -> bs4.PageElement:
        """Convert the element to a corresponding `BeautifulSoup` object."""

    @abc.abstractmethod
    def _eq(self, other: Element, /) -> bool: ...

    @property
    def parents(self) -> Generator[Tag]:
        curr = self.parent

        while curr is not None:
            yield curr
            curr = curr.parent

    @override
    def __eq__(self, other: object) -> bool:
        if not utils.basic_compare(other, self=self):
            return False

        return self._eq(other)


class Tag(
    Element, groups.Core, groups.Presentation, metaclass=abc.ABCMeta
):
    """A tag.

    A tag is an element that has a name, a set of attributes and (optionally)
    one or more elements as children.

    Tags can have standard attributes (for example, `id`, `class`, `color`) and
    non-standard user-defined attributes.

    Example: `<rect id="foo" class="bar" color="red" />`
    ```
    """

    model_config = pydantic.ConfigDict(
        extra="allow",
        alias_generator=pydantic.AliasGenerator(
            validation_alias=lambda name: pydantic.AliasChoices(
                name,
                attr_names.ATTR_NAME_TO_NORMALIZED.inverse.get(name, name),
            ),
            serialization_alias=(
                lambda name: attr_names.ATTR_NAME_TO_NORMALIZED.inverse.get(
                    name, name
                )
            ),
        ),
    )

    prefix: str | None = None

    __children: list[Element] = pydantic.PrivateAttr(default_factory=list)

    @pydantic.model_validator(mode="after")
    def __validate_extra(self) -> Tag:  # pyright: ignore[reportUnusedFunction]
        # model_extra cannot be None because extra is set to "allow"
        assert self.model_extra is not None, "model_extra is None"

        for key, value in self.model_extra.items():
            if not isinstance(value, str):
                msg = (
                    f"Non-standard attribute {key!r} must be of type str,"
                    f" got {type(value)}"
                )
                raise TypeError(msg)

        return self

    def extra_attrs(self) -> Mapping[str, str]:
        assert self.model_extra is not None, "model_extra is None"
        return self.model_extra

    def standard_attrs(self) -> Mapping[attr_names.AttributeName, object]:
        dump = self.model_dump(by_alias=True, exclude_none=True)

        return {
            attr: getattr(self, attr_names.ATTR_NAME_TO_NORMALIZED[attr])
            for key, _ in dump.items()
            if (attr := cast(attr_names.AttributeName, key))
            in attr_names.ATTR_NAME_TO_NORMALIZED
        }

    def all_attrs(self) -> Mapping[str, object]:
        standard = cast(Mapping[str, object], self.standard_attrs())
        extra = self.extra_attrs()

        return {**standard, **extra}

    @override
    def _eq(self, other: Element) -> bool:
        return (
            isinstance(other, Tag)
            and self.prefix == other.prefix
            and self.all_attrs() == other.all_attrs()
            and self.num_children == other.num_children
            and all(
                c1 == c2
                for c1, c2 in zip(
                    self.children, other.children, strict=True
                )
            )
        )

    @property
    def children(self) -> Generator[Element]:
        yield from self.__children

    @property
    def descendants(self) -> Generator[Element]:
        queue = collections.deque(self.children)

        while queue:
            child = queue.popleft()
            yield child

            if isinstance(child, Tag):
                queue.extend(child.children)

    @property
    def next_siblings(self) -> Generator[Element]:
        if self.parent is None:
            return

        should_yield = False

        for sibling in self.parent.children:
            if should_yield:
                yield sibling
            elif sibling is self:
                should_yield = True

    @property
    def prev_siblings(self) -> Generator[Element]:
        if self.parent is None:
            return

        for sibling in self.parent.children:
            if sibling is self:
                return

            yield sibling

    @property
    def siblings(self) -> Generator[Element]:
        yield from self.prev_siblings
        yield from self.next_siblings

    def add_child(
        self, child: Element, /, *, index: int | None = None
    ) -> Self:
        if child is self:
            raise ValueError("Cannot add a tag as a child of itself.")

        if child.parent is not None:
            raise ValueError(
                "Cannot add a child that already has a parent."
            )

        if index is None:
            self.__children.append(child)
        else:
            self.__children.insert(index, child)

        child.parent = self

        return self

    def add_children(self, *children: Element) -> Self:
        for child in children:
            self.add_child(child)

        return self

    def remove_child(self, child: Element, /) -> Element:
        for i, elem in enumerate(self.children):
            if elem is child:
                return self.pop_child(i)

        raise ValueError("Child not found")

    def pop_child(self, index: SupportsIndex = -1, /) -> Element:
        child = self.__children.pop(index)
        child.parent = None

        return child

    def clear_children(self) -> None:
        for child in self.__children:
            child.parent = None

        self.__children.clear()

    def get_child(self, index: SupportsIndex = -1, /) -> Element:
        return self.__children[index]

    def get_child_index(
        self,
        child: Element,
        /,
        start: SupportsIndex = 0,
        stop: SupportsIndex = sys.maxsize,
    ) -> int:
        return self.__children.index(child, start, stop)

    @override
    def to_beautifulsoup_object(self) -> bs4.Tag:
        tag = bs4.Tag(
            name=tag_name(self),
            can_be_empty_element=True,
            prefix=self.prefix,
            is_xml=True,
        )

        tag.can_be_empty_element = len(self.__children) == 0

        for key, value in self.all_attrs().items():
            tag[key] = serialize.serialize_attr(key, value)

        for child in self.children:
            tag.append(child.to_beautifulsoup_object())

        if tag_name(self) == "svg":
            formatter = serialize.get_current_formatter()

            if formatter.xmlns == "always":
                tag["xmlns"] = constants.SVG_XMLNS
            elif formatter.xmlns == "never":
                del tag["xmlns"]

        return tag

    def decompose_transform_origin(self) -> None:
        """Decompose the `transform-origin` attribute into `transform`.

        This function replaces the `transform-origin` attribute with a pair of
        `Translate` transformations in the `transform` attribute. The resulting
        `transform` attribute looks like this:

        ```
        self.transform = [
            Translate(tx, ty),
            *self.transform,
            Translate(-tx, -ty),
        ]
        ```

        If the `transform-origin` attribute is not set, this function does
        nothing.

        Raises:
            SvgTransformOriginError: If the value of the `transform-origin`
                attribute cannot be decomposed.

        """
        if self.transform_origin is None:
            return

        if (
            isinstance(self.transform_origin, tuple)
            and isinstance(self.transform_origin[0], length.Length)
            and isinstance(self.transform_origin[1], length.Length)
        ):
            tx = float(self.transform_origin[0])
            ty = float(self.transform_origin[1])

            if not self.transform:
                self.transform = []

            self.transform.insert(0, transform.Translate(tx, ty))
            self.transform.append(transform.Translate(-tx, -ty))

            self.transform_origin = None
        else:
            raise errors.SvgTransformOriginError(self.transform_origin)

    def __reify_this(self, *, limit: int = sys.maxsize) -> None:
        if limit < 0:
            raise ValueError("Limit must be a positive integer")

        self.decompose_transform_origin()

        if not self.transform:
            return

        transform.decompose_matrices(transform=self.transform)

        reified = 0
        i = 0

        while reified < limit and i < len(self.transform):
            if not isinstance(self.transform[i], transform.Reifiable):
                i += 1
                continue

            try:
                # move the transformation to the end of the list where it can
                # be directly applied to the elementf
                _move_transformation_to_end(self.transform, i)
                transformation = self.transform.pop()

                _apply_transformation(self, transformation)

                for child in self.find_all(recursive=False):
                    if child.transform is None:
                        child.transform = []

                    # decompose transform-origin before we prepend
                    child.decompose_transform_origin()
                    child.transform.insert(0, transformation)

                    child.reify(
                        limit=1,
                        recursive=False,
                        remove_transform_list_if_empty=False,
                    )
            except (ValueError, errors.SvgTransformSwapError) as e:
                raise errors.SvgReifyError from e

            reified += 1

    def reify(
        self,
        *,
        limit: int = sys.maxsize,
        recursive: bool = True,
        remove_transform_list_if_empty: bool = True,
    ) -> None:
        """Apply transformations defined by the `transform` attribute.

        This method takes the transformations defined by the `transform`
        attribute and applies them directly to the coordinate, length, and
        other attributes of the element. The transformations are applied in
        the order in which they are defined. The result of this operation
        should be a visually identical element with the `transform` attribute
        reduced or removed (depending on the `limit` parameter).

        Only `Translate` and `Scale` transformations are can be reified. If
        the `transform` attribute contains other transformations, their
        parameters are adjusted so that `Translate` and `Scale` transformations
        can be applied. Unsupported transformations are ignored.

        If all transformations are successfully applied, the `transform`
        attribute is removed from the element.

        All length values in the element must be convertible to user units.

        Args:
            limit: The maximum number of transformations to apply. If the
                `transform` attribute contains more transformations than the
                limit, the remaining transformations are kept in the attribute
                and not applied. The limit must be a positive integer and is
                applied on a per-element basis.
            recursive: If `True`, the method is called recursively on all
                child elements that support reification.
            remove_transform_list_if_empty: If `True`, the `transform`
                attribute is set to `None` if the list is empty after
                reification.

        Raises:
            ValueError: If the limit is not a positive integer.
            SvgReifyError: If the `transform` attribute cannot be reified.
            SvgUnitConversionError: If a length value cannot be converted to
                user units.
            SvgTransformOriginError: If the value of the `transform-origin`
                attribute is unsupported.

        Examples:
            >>> from svglab import Rect, Length, Translate
            >>> rect = Rect(
            ...     x=Length(10),
            ...     y=Length(20),
            ...     width=Length(20),
            ...     height=Length(40),
            ...     transform=[Translate(5, 5)],
            ... )
            >>> rect.reify()
            >>> rect.x
            Length(value=15.0, unit=None)
            >>> rect.y
            Length(value=25.0, unit=None)
            >>> rect.width
            Length(value=20.0, unit=None)
            >>> rect.height
            Length(value=40.0, unit=None)
            >>> rect.transform is None
            True

        """
        self.__reify_this(limit=limit)

        if remove_transform_list_if_empty and not self.transform:
            self.transform = None

        if recursive:
            for child in self.find_all(recursive=False):
                child.reify(
                    limit=limit,
                    recursive=True,
                    remove_transform_list_if_empty=remove_transform_list_if_empty,
                )

    @overload
    def find_all(self, /, *, recursive: bool = True) -> Generator[Tag]: ...

    @overload
    def find_all(
        self, *tags: type[_T_tag], recursive: bool = True
    ) -> Generator[_T_tag]: ...

    @overload
    def find_all(
        self, *tags: type[Tag] | names.TagName, recursive: bool = True
    ) -> Generator[Tag]: ...

    def find_all(
        self, *tags: type[Tag] | names.TagName, recursive: bool = True
    ) -> Generator[Tag]:
        """Find all tags that match the given search criteria.

        Args:
        tags: The tags to search for. Can be tag names or tag classes.
            If no search criteria are provided, all tags are returned.
        recursive: If `False`, only search the direct children of the tag,
        otherwise search all descendants.

        Returns:
        An iterator over all tags that match the search criteria.

        Examples:
        >>> from svglab import G, Rect
        >>> g = G().add_children(Rect(), G().add_child(Rect()))
        >>> list(g.find_all("rect"))
        [Rect(), Rect()]
        >>> list(g.find_all(G))
        [G(children=[Rect()])]
        >>> list(g.find_all(Rect, recursive=False))
        [Rect()]
        >>> list(g.find_all(G, "rect"))
        [Rect(), G(children=[Rect()]), Rect()]

        """
        for child in self.descendants if recursive else self.children:
            if isinstance(child, Tag) and (
                len(tags) == 0
                or any(_match_tag(child, search=tag) for tag in tags)
            ):
                yield child

    @overload
    def find(
        self, *tags: type[_T_tag], recursive: bool = True
    ) -> _T_tag: ...

    @overload
    def find(
        self, *tags: type[Tag] | names.TagName, recursive: bool = True
    ) -> Tag: ...

    @overload
    def find(
        self,
        *tags: type[_T_tag],
        recursive: bool = True,
        default: _T = _EMPTY_PARAM,
    ) -> _T_tag | _T: ...

    @overload
    def find(
        self,
        *tags: type[Tag] | names.TagName,
        recursive: bool = True,
        default: _T = _EMPTY_PARAM,
    ) -> Tag | _T: ...

    def find(
        self,
        *tags: type[Tag] | names.TagName,
        recursive: bool = True,
        default: _T = _EMPTY_PARAM,
    ) -> Tag | _T:
        """Find the first tag that matches the given search criteria.

        Args:
        tags: The tags to search for. Can be tag names or tag classes.
        recursive: If `False`, only search the direct children of the tag,
        otherwise search all descendants.
        default: The default value to return if no tag matches the search
        criteria.

        Returns:
        The first tag that matches the search criteria.

        Raises:
        SvgElementNotFoundError: If no tag matches the search criteria and no
        default value is provided.

        Examples:
        >>> from svglab import G, Rect
        >>> g = G().add_children(
        ...     Rect(id="foo"), G().add_child(Rect(id="bar"))
        ... )
        >>> g.find("rect")
        Rect(id='foo')
        >>> g.find(G)
        G(children=[Rect(id='bar')])
        >>> g.find("circle", default=None) is None
        True

        """
        try:
            return next(self.find_all(*tags, recursive=recursive))
        except StopIteration as e:
            if default is not _EMPTY_PARAM:
                return default

            msg = f"Unable to find tag by search criteria: {tags}"
            raise errors.SvgElementNotFoundError(msg) from e

    @property
    def num_children(self) -> int:
        return len(self.__children)

    def has_children(self) -> bool:
        return self.num_children > 0

    def __getitem__(self, key: str) -> str:
        assert self.model_extra is not None, "model_extra is None"
        value: str = self.model_extra[key]
        return value

    def __setitem__(self, key: str, value: str) -> None:
        assert self.model_extra is not None, "model_extra is None"
        self.model_extra[key] = value

    def __delitem__(self, key: str) -> None:
        assert self.model_extra is not None, "model_extra is None"
        del self.model_extra[key]

    @reprlib.recursive_repr()
    @override
    def __repr__(self) -> str:
        name = type(self).__name__
        attrs = dict(self.all_attrs())

        if self.__children:
            attrs["children"] = list(self.children)

        attr_repr = ", ".join(
            f"{key}={value!r}" for key, value in attrs.items()
        )

        return f"{name}({attr_repr})"


def _match_tag(tag: Tag, /, *, search: type[Tag] | names.TagName) -> bool:
    """Check if a tag matches the given search criteria.

    Args:
    tag: The tag to check.
    search: The search criteria. Can be a tag name or a tag class.

    Returns:
    `True` if the tag matches the search criteria, `False` otherwise.

    Examples:
    >>> from svglab import Rect
    >>> rect = Rect()
    >>> _match_tag(rect, search="rect")
    True
    >>> _match_tag(rect, search=Rect)
    True
    >>> _match_tag(rect, search="circle")
    False

    """
    if isinstance(search, type):
        return isinstance(tag, search)

    return tag_name(tag) == search


class TextElement(Element, metaclass=abc.ABCMeta):
    """The base class of text-based elements.

    Text-based elements are elements that are represented by a single string.

    Common examples of text-based elements in XML are:
    - `CDATA` sections (`CData`)
    - comments (`Comment`)
    - text (`Text`)
    - processing instructions (for example, `<?xml version="1.0"?>`)
    - Document Type Definitions (DTDs; for example, `<!DOCTYPE html>`)
    """

    content: str = pydantic.Field(frozen=True, min_length=1)

    @override
    def _eq(self, other: Element) -> bool:
        return (
            isinstance(other, TextElement)
            and self.content == other.content
        )

    @override
    def __repr__(self) -> str:
        name = type(self).__name__
        return f"{name}({self.content!r})"



import base64
import contextlib
import os
import pathlib

import PIL.Image
from typing_extensions import final, overload

from svglab import graphics, protocols, serialize, utils
from svglab.attrparse import transform
from svglab.attrs import groups, regular
from svglab.elements import traits


@final
class Svg(
    groups.ConditionalProcessing,
    groups.DocumentEvents,
    regular.BaseProfile,
    regular.Class,
    regular.ContentScriptType,
    regular.ContentStyleType,
    regular.ExternalResourcesRequired,
    regular.Height,
    regular.PreserveAspectRatio,
    regular.Style,
    regular.Version,
    regular.ViewBox,
    regular.Width,
    regular.XCoordinate,
    regular.Xmlns,
    regular.YCoordinate,
    regular.ZoomAndPan,
    traits.StructuralElement,
    traits.ContainerElement,
):
    @overload
    def save(
        self,
        path: str | os.PathLike[str],
        /,
        *,
        pretty: bool = True,
        trailing_newline: bool = True,
        formatter: serialize.Formatter | None = None,
    ) -> None: ...

    @overload
    def save(
        self,
        file: protocols.SupportsWrite[str],
        /,
        *,
        pretty: bool = True,
        trailing_newline: bool = True,
        formatter: serialize.Formatter | None = None,
    ) -> None: ...

    def save(
        self,
        path_or_file: str
        | os.PathLike[str]
        | protocols.SupportsWrite[str],
        /,
        *,
        pretty: bool = True,
        trailing_newline: bool = True,
        formatter: serialize.Formatter | None = None,
    ) -> None:
        """Convert the SVG document fragment to XML and write it to a file.

        Args:
        path_or_file: The path to the file to save the XML to,
        or a file-like object.
        pretty: Whether to produce pretty-printed XML.
        indent: The number of spaces to indent each level of the document.
        trailing_newline: Whether to add a trailing newline to the file.
        formatter: The formatter to use for serialization.

        Examples:
        >>> import sys
        >>> from svglab import Rect, Svg
        >>> svg = Svg(id="foo").add_child(Rect())
        >>> formatter = serialize.Formatter(indent=4)
        >>> svg.save(
        ...     sys.stdout,
        ...     pretty=True,
        ...     trailing_newline=False,
        ...     formatter=formatter,
        ... )
        <svg id="foo">
            <rect/>
        </svg>

        """
        with contextlib.ExitStack() as stack:
            output = self.to_xml(pretty=pretty, formatter=formatter)
            file: protocols.SupportsWrite[str]

            match path_or_file:
                case str() | os.PathLike() as path:
                    file = stack.enter_context(
                        pathlib.Path(path).open("w")
                    )
                case protocols.SupportsWrite() as file:
                    pass

            file.write(output)

            if trailing_newline:
                file.write("\n")

    def set_viewbox(
        self, viewbox: tuple[float, float, float, float]
    ) -> None:
        """Set a new value for the `viewBox` attribute.

        This method sets a new value for the `viewBox` attribute and scales and
        translates the SVG content so that the visual representation of the SVG
        remains unchanged.

        If the `viewBox` is not set, the method uses the `width` and `height`
        attributes to calculate the initial viewBox. If the `width`
        and `height` attributes are not set, the method raises an exception.

        The new `viewBox` must have the same aspect ratio as the old `viewBox`.
        If the aspect ratios differ, the method raises an exception.

        Any attributes of type `Length` in the SVG must be convertible to
        user units. If an attribute is not convertible, the method raises an
        exception.

        Args:
        viewbox: A tuple of four numbers representing the new viewBox.

        Raises:
        ValueError: If `viewBox` is not set and `width` and `height` are not
            set or if the aspect ratios of the old and new viewBox differ.
        SvgUnitConversionError: If an attribute is not convertible to user
            units.

        """
        if self.viewBox is None:
            if self.width is None or self.height is None:
                raise ValueError(
                    "Either viewBox or width and height must be set"
                )
            old_viewbox = (0, 0, float(self.width), float(self.height))
        else:
            old_viewbox = self.viewBox

        old_min_x, old_min_y, old_width, old_height = old_viewbox
        min_x, min_y, width, height = viewbox

        tx = min_x - old_min_x
        ty = min_y - old_min_y

        sx = width / old_width
        sy = height / old_height

        if not utils.is_close(sx, sy):
            raise ValueError("Aspect ratios of old and new viewBox differ")

        # skip self; this can be done in a single for loop because the
        # SVG is a tree (probably)
        for child in self.find_all(recursive=False):
            # this is normally done in the reify method, but we need to do it
            # before we prepend the new transformations
            child.decompose_transform_origin()

            child.transform = [
                transform.Translate(tx, ty),
                transform.Scale(sx, sy),
                *(child.transform or []),
            ]

            child.reify(limit=2, recursive=False)

        self.viewBox = viewbox

    def render(
        self, *, width: float | None = None, height: float | None = None
    ) -> PIL.Image.Image:
        """Render an SVG document fragment into a Pillow image.

        If the width and height are not specified, the dimensions of the SVG
        element are used. If only one dimension is specified, the other
        dimension is calculated in a way that preserves the aspect ratio set
        in the SVG element. If both dimensions are specified, the aspect ratio
        must match the aspect ratio defined by `width` and `height` attributes
        of the SVG element.

        Args:
        svg: The SVG document fragment to render.
        width: The width of the rendered image, in pixels. If `None`, the width
            attribute of the SVG element is used.
        height: The height of the rendered image, in pixels. If `None`, the
            height attribute of the SVG element is used.

        Returns:
        The rendered image.

        """
        return graphics.render(self, width=width, height=height)

    def show(self) -> None:
        """Render this SVG document fragment and display it on the screen.

        See `PIL.Image.Image.show` for more information.

        """
        self.render().show()

    def to_data_uri(self) -> str:
        """Convert the SVG document fragment to a base64-encoded data URI."""
        xml = self.to_xml(formatter=serialize.MINIMAL_FORMATTER)
        b64 = base64.b64encode(xml.encode()).decode()

        return f"data:image/svg+xml;base64,{b64}"



import abc

from typing_extensions import Protocol

from svglab import graphics, models
from svglab.attrparse import d
from svglab.attrs import groups, regular
from svglab.elements import common


# common attributes are defined directly on the Tag class
class Element(common.Tag):
    pass


class _GraphicalOperations(Element):
    def get_bbox(
        self, *, visible_only: bool = False
    ) -> graphics.BBox | None:
        """Compute the bounding box of this element.

        The bounding box is the smallest rectangle that contains the entire
        element. If the element is not visible, the bounding box is `None`.

        Args:
            visible_only: If `True`, only the visible parts of the element are
                considered when computing the bounding box. If `False`, the
                bounding box includes all parts of the element (even if they
                are transparent).

        Returns:
            The bounding box of the element, or `None` if the element is not
            visible. The bounding box is a tuple of the form `(x_min, y_min,
            x_max, y_max)`.

        """
        return (
            graphics.visible_bbox(self)
            if visible_only
            else graphics.bbox(self)
        )

    def get_mask(
        self,
        *,
        visible_only: bool = False,
        width: float | None = None,
        height: float | None = None,
    ) -> graphics.Mask:
        """Create a mask of this element.

        A mask is a 2D boolean array with `True` values where the tag is
        located (or visible) in the rendered SVG and `False` values elsewhere.

        Args:
            tag: The tag to create a mask for.
            visible_only: If `True`, only the visible parts of the tag are
                included in the mask. If `False`, the mask includes all parts
                of the tag (even if they are transparent).
            width: The width of the mask. If `None`, the width of the root
                SVG tag in the tree is used.
            height: The height of the mask. If `None`, the height of the root
                SVG tag in the tree is used.

        Returns:
            A 2D boolean array representing the mask of the tag.

        """
        return (
            graphics.visible_mask(self, width=width, height=height)
            if visible_only
            else graphics.mask(self, width=width, height=height)
        )


class GraphicsElement(
    _GraphicalOperations,
    groups.GraphicalEvents,
    common.StrokeWidthScaled,
    Element,
):
    pass


class Shape(regular.PathLength, GraphicsElement):
    def set_path_length(self, value: float) -> None:
        """Set a new value for the `pathLength` attribute.

        This method sets a new value for the `pathLength` attribute and scales
        the shape's `stroke-dasharray` and `stroke-dashoffset` attributes
        so that the visual appearance of the shape remains unchanged.

        The shape must have the `pathLength` attribute defined. The `Length`
        values of the scaled attributes must be either in percentages (`%`) or
        convertible to user units. Percentage values are not scaled.

        Args:
            value: The new value for the `pathLength` attribute. Must be
                positive.

        Raises:
            ValueError: If the `value` is not positive.
            RuntimeError: If the current path length is `None`.
            SvgUnitConversionError: If the attribute values cannot be converted
                to user units.

        """
        if value <= 0:
            raise ValueError("Path length must be positive")

        if self.pathLength is None:
            raise RuntimeError("Current pathLength must not be None")

        ratio = value / self.pathLength
        common.scale_distance_along_a_path_attrs(self, ratio)

        self.pathLength = value


class _PathLike(Protocol):
    d: models.Attr[d.D]


class BasicShape(Shape, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def to_d(self) -> d.D:
        """Convert this basic shape into path data.

        The resulting path data produce the same visual result as the original
        basic shape.

        Returns:
            A `D` instance representing the path data.

        """
        ...

    @abc.abstractmethod
    def to_path(self) -> _PathLike:
        """Convert this basic shape into a `Path` element.

        The resulting `Path` element's path data produce the same visual result
        as the original basic shape. The `Path` element will have the same
        attributes as the original basic shape.

        Returns:
            A `Path` element representing the basic shape.

        """
        ...


class AnimationElement(
    groups.AnimationEvents, groups.AnimationTiming, Element
):
    pass


class ContainerElement(
    _GraphicalOperations, groups.GraphicalEvents, Element
):
    pass


class DescriptiveElement(Element):
    pass


class FilterPrimitiveElement(groups.FilterPrimitives, Element):
    pass


class GradientElement(Element):
    pass


class GraphicsReferencingElement(Element):
    pass


class LightSourceElement(Element):
    pass


class StructuralElement(Element):
    pass


class TextContentElement(GraphicsElement):
    pass


class TextContentChildElement(Element):
    pass


class TextContentBlockElement(Element):
    pass



from __future__ import annotations

import abc
import contextlib
from collections.abc import Generator, Iterable, MutableSequence

import lark
import pydantic
import pydantic_core
from typing_extensions import (
    Annotated,
    Final,
    Literal,
    Protocol,
    Self,
    SupportsIndex,
    TypeAlias,
    cast,
    final,
    overload,
    override,
    runtime_checkable,
)

from svglab import errors, mixins, protocols, serialize, utils
from svglab.attrparse import parse, point, transform


_Flag: TypeAlias = Literal["0", "1"]


@pydantic.dataclasses.dataclass
class _PathCommandBase:
    pass


@runtime_checkable
class _HasEnd(Protocol):
    end: point.Point


class _PhysicalPathCommand(
    _PathCommandBase,
    transform.PointAddSubWithTranslateRMatmul,
    metaclass=abc.ABCMeta,
):
    pass


@final
@pydantic.dataclasses.dataclass
class ClosePath(_PathCommandBase):
    pass


@pydantic.dataclasses.dataclass
class LineTo(_HasEnd, _PhysicalPathCommand):
    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(end=other @ self.end)


@final
@pydantic.dataclasses.dataclass
class HorizontalLineTo(_PhysicalPathCommand):
    x: float

    def to_line(self) -> LineTo:
        return LineTo(end=point.Point(self.x, 0))

    @overload
    def __rmatmul__(
        self, other: transform.Translate | transform.Scale
    ) -> Self: ...

    @overload
    def __rmatmul__(self, other: transform.TransformFunction) -> Self: ...

    @override
    def __rmatmul__(
        self, other: transform.TransformFunction
    ) -> Self | LineTo:
        if isinstance(other, transform.Translate | transform.Scale):
            x, _ = other @ point.Point(self.x, 0)

            return type(self)(x=x)

        return other @ self.to_line()


@final
@pydantic.dataclasses.dataclass
class VerticalLineTo(_PhysicalPathCommand):
    y: float

    def to_line(self) -> LineTo:
        return LineTo(end=point.Point(0, self.y))

    @overload
    def __rmatmul__(
        self, other: transform.Translate | transform.Scale
    ) -> Self: ...

    @overload
    def __rmatmul__(self, other: transform.TransformFunction) -> Self: ...

    @override
    def __rmatmul__(
        self, other: transform.TransformFunction
    ) -> Self | LineTo:
        if isinstance(other, transform.Translate | transform.Scale):
            _, y = other @ point.Point(0, self.y)

            return type(self)(y=y)

        return other @ self.to_line()


@final
@pydantic.dataclasses.dataclass
class SmoothQuadraticBezierTo(_HasEnd, _PhysicalPathCommand):
    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(end=other @ self.end)


@final
@pydantic.dataclasses.dataclass
class SmoothCubicBezierTo(_HasEnd, _PhysicalPathCommand):
    control2: point.Point
    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(
            control2=other @ self.control2, end=other @ self.end
        )


@final
@pydantic.dataclasses.dataclass
class MoveTo(_HasEnd, _PhysicalPathCommand):
    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(end=other @ self.end)


@final
@pydantic.dataclasses.dataclass
class QuadraticBezierTo(_HasEnd, _PhysicalPathCommand):
    control: point.Point
    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(
            control=other @ self.control, end=other @ self.end
        )


@final
@pydantic.dataclasses.dataclass
class CubicBezierTo(_HasEnd, _PhysicalPathCommand):
    control1: point.Point
    control2: point.Point
    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(
            control1=other @ self.control1,
            control2=other @ self.control2,
            end=other @ self.end,
        )


@final
@pydantic.dataclasses.dataclass
class ArcTo(_HasEnd, _PhysicalPathCommand):
    radii: point.Point
    angle: float
    large: bool
    sweep: bool
    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        radii = self.radii
        angle = self.angle
        end = self.end

        match other:
            case transform.Translate():
                end = other @ end
            case transform.Scale():
                radii = other @ radii
                end = other @ end
            case transform.Rotate(a):
                angle += a
            case _:
                msg = f"Unsupported transform: {other}"
                raise NotImplementedError(msg)

        return type(self)(
            radii=radii,
            angle=angle,
            large=self.large,
            sweep=self.sweep,
            end=end,
        )


PathCommand: TypeAlias = (
    ArcTo
    | ClosePath
    | CubicBezierTo
    | HorizontalLineTo
    | LineTo
    | MoveTo
    | QuadraticBezierTo
    | SmoothCubicBezierTo
    | SmoothQuadraticBezierTo
    | VerticalLineTo
)


def _get_end(d: D, command: PathCommand) -> point.Point:
    """Get the end point of a path command.

    For commands that have a set end point, this function returns the end
    point. For commands that do not have a set end point, this function
    calculates the end point based on the previous command.

    Args:
        d: The path containing the command.
        command: The command to get the end point of.

    Returns:
        The end point of the command.

    Examples:
    >>> d = D.from_str("M 10,10 H 100 V 100 Z")
    >>> _get_end(d, d[0])
    Point(x=10.0, y=10.0)
    >>> _get_end(d, d[1])
    Point(x=100.0, y=10.0)

    """
    match command:
        case ClosePath():
            return _get_end(d, utils.prev(d, command))
        case HorizontalLineTo(x=x):
            end = _get_end(d, utils.prev(d, command))
            return point.Point(x, end.y)
        case VerticalLineTo(y=y):
            end = _get_end(d, utils.prev(d, command))
            return point.Point(end.x, y)
        case _:
            return command.end


def _quadratic_control(
    d: D, command: SmoothQuadraticBezierTo
) -> point.Point:
    """Compute the control point for a smooth quadratic Bézier command (`T`).

    The control point is calculated based on the previous command. If the
    previous command is a quadratic Bézier command (`Q`), the control point
    is the reflection of the previous control point across the end point of
    the previous command. If the previous command is of any other type, the
    control point is coincident with the end point of the previous command.

    Args:
        d: The path containing the command.
        command: The smooth quadratic Bézier command.

    Returns:
        The control point for the command.

    Examples:
    >>> d = D.from_str("M 0,0 Q 20,0 20,20 T 40,40")
    >>> _quadratic_control(d, d[2])
    Point(x=20.0, y=40.0)

    """
    prev = utils.prev(d, command)
    end = _get_end(d, prev)

    match prev:
        case QuadraticBezierTo(control=control):
            pass
        case SmoothQuadraticBezierTo():
            control = _quadratic_control(d, prev)
        case _:
            return end

    return control.line_reflect(end)


def _cubic_control(d: D, command: SmoothCubicBezierTo) -> point.Point:
    """Compute the first control point for a smooth cubic Bézier command (`S`).

    The control point is calculated based on the previous command. If the
    previous command is a cubic Bézier command (`C`), the control point is the
    reflection of the second control point of the previous command across the
    end point of the previous command. If the previous command is of any other
    type, the control point is coincident with the end point of the previous
    command.

    Args:
        d: The path containing the command.
        command: The smooth cubic Bézier command.

    Returns:
        The first control point for the command.

    Examples:
    >>> d = D.from_str("M 0,0 C 20,0 20,20 40,40 S 100,100 50,50")
    >>> _cubic_control(d, d[2])
    Point(x=60.0, y=60.0)

    """
    prev = utils.prev(d, command)
    end = _get_end(d, prev)

    if isinstance(prev, CubicBezierTo | SmoothCubicBezierTo):
        return prev.control2.line_reflect(end)

    return end


def _relativize(d: D) -> D:
    """Recompute the coordinates of path commands as if they were relative.

    This function takes a path and recomputes the coordinates of the path
    commands as if they were relative. This is done by subtracting the
    current position from the coordinates of each command. The resuting path
    may be used for serialization with relative coordinates.

    Args:
        d: The path to relativize.

    Returns:
        A new `D` instance with the coordinates of the path commands recomputed
        as if they were relative.

    Examples:
    >>> d = D.from_str("M 10,10 L 100,100")
    >>> _relativize(d)
    D(MoveTo(end=Point(x=10.0, y=10.0)), LineTo(end=Point(x=90.0, y=90.0)))

    """
    result = D()
    pos = point.Point.zero()

    for command in d:
        if isinstance(command, _PhysicalPathCommand):
            # TODO: this cast can be removed fairly easily
            result.append(cast(PathCommand, command - pos))
        else:
            result.append(command)

        pos = _get_end(d, command)

    return result


def _can_use_implicit_command(
    current: PathCommand, /, *, prev: PathCommand | None
) -> bool:
    """Determine whether a command can be serialized implicitly.

    A command can be serialized implicitly if the previous command is of the
    same type as the current command, or if the previous command is a `MoveTo`
    command and the current command is a `LineTo` command.

    Args:
        prev: The previous command.
        current: The current command.

    Returns:
        `True` if the command can be serialized implicitly, `False` otherwise.

    Examples:
    >>> _can_use_implicit_command(MoveTo(point.Point(10, 10)), prev=None)
    False
    >>> _can_use_implicit_command(
    ...     LineTo(point.Point(100, 100)), prev=MoveTo(point.Point(10, 10))
    ... )
    True
    >>> _can_use_implicit_command(
    ...     LineTo(point.Point(100, 100)), prev=LineTo(point.Point(10, 10))
    ... )
    True

    """
    return type(prev) is type(current) or (
        isinstance(prev, MoveTo) and isinstance(current, LineTo)
    )


def _add_command(
    d: D, command: PathCommand, *, relative: bool = False
) -> None:
    if relative and isinstance(command, _PhysicalPathCommand):
        # if there is no previous command, just do nothing
        with contextlib.suppress(IndexError):
            command += _get_end(d, d[-1])

    d.append(command)


@final
class D(
    MutableSequence[PathCommand],
    mixins.CustomModel,
    transform.PointAddSubWithTranslateRMatmul,
    protocols.CustomSerializable,
):
    """A class representing the `d` attribute of a path element.

    The `d` attribute is used to define a path in SVG. This class provides
    methods for building a path by adding commands. `D` is also
    a `MutableSequence` of `PathCommand` instances.

    Args:
        iterable: An iterable of `PathCommand` instances (for example,
        another `D` instance).
        start: The starting point of the path. If `start` is not `None`, a
        `MoveTo` command is automatically added to the path, moving the "pen"
        to the starting point.

    Examples:
    >>> d = (
    ...     D()
    ...     .move_to(point.Point(10, 10))
    ...     .line_to(point.Point(100, 100), relative=True)
    ... )
    >>> d
    D(MoveTo(end=Point(x=10.0, y=10.0)), LineTo(end=Point(x=110.0, y=110.0)))
    >>> len(d)
    2
    >>> bool(d)
    True
    >>> d[0]
    MoveTo(end=Point(x=10.0, y=10.0))
    >>> d.pop()
    LineTo(end=Point(x=110.0, y=110.0))
    >>> d.close()
    D(MoveTo(end=Point(x=10.0, y=10.0)), ClosePath())

    """

    def __add(
        self, command: PathCommand, /, *, relative: bool = False
    ) -> Self:
        _add_command(self, command, relative=relative)

        return self

    def __init__(
        self,
        iterable: Iterable[PathCommand] = (),
        /,
        *,
        start: point.Point | None = None,
    ) -> None:
        self.__commands: Final[list[PathCommand]] = []

        if start is not None:
            self.move_to(start)

        for command in iterable:
            self.__add(command)

    @override
    def insert(self, index: SupportsIndex, value: PathCommand) -> None:
        if utils.is_first_index(self, index) and not isinstance(
            value, MoveTo
        ):
            raise errors.SvgPathMissingMoveToError

        self.__commands.insert(index, value)

    def move_to(
        self, end: point.Point, /, *, relative: bool = False
    ) -> Self:
        return self.__add(MoveTo(end=end), relative=relative)

    def line_to(
        self, end: point.Point, /, *, relative: bool = False
    ) -> Self:
        return self.__add(LineTo(end=end), relative=relative)

    def horizontal_line_to(
        self,
        x: protocols.SupportsFloatOrIndex,
        /,
        *,
        relative: bool = False,
    ) -> Self:
        return self.__add(HorizontalLineTo(x=float(x)), relative=relative)

    def vertical_line_to(
        self,
        y: protocols.SupportsFloatOrIndex,
        /,
        *,
        relative: bool = False,
    ) -> Self:
        return self.__add(VerticalLineTo(y=float(y)), relative=relative)

    def quadratic_bezier_to(
        self,
        control: point.Point,
        end: point.Point,
        /,
        *,
        relative: bool = False,
    ) -> Self:
        return self.__add(
            QuadraticBezierTo(control=control, end=end), relative=relative
        )

    def cubic_bezier_to(
        self,
        control1: point.Point,
        control2: point.Point,
        end: point.Point,
        *,
        relative: bool = False,
    ) -> Self:
        return self.__add(
            CubicBezierTo(control1=control1, control2=control2, end=end),
            relative=relative,
        )

    def arc_to(
        self,
        radii: point.Point,
        angle: float,
        end: point.Point,
        *,
        large: bool,
        sweep: bool,
        relative: bool = False,
    ) -> Self:
        return self.__add(
            ArcTo(
                radii=radii, angle=angle, large=large, sweep=sweep, end=end
            ),
            relative=relative,
        )

    def smooth_quadratic_bezier_to(
        self, end: point.Point, /, *, relative: bool = False
    ) -> Self:
        return self.__add(
            SmoothQuadraticBezierTo(end=end), relative=relative
        )

    def smooth_cubic_bezier_to(
        self,
        control2: point.Point,
        end: point.Point,
        /,
        *,
        relative: bool = False,
    ) -> Self:
        return self.__add(
            SmoothCubicBezierTo(control2=control2, end=end),
            relative=relative,
        )

    def close(self) -> Self:
        return self.__add(ClosePath())

    @classmethod
    def from_str(cls, text: str) -> Self:
        d = parse.parse(text, grammar="d.lark", transformer=_Transformer())

        assert isinstance(d, cls), f"Expected {cls}, got {type(d)}"
        return d

    @classmethod
    def _validate(
        cls, value: object, info: pydantic_core.core_schema.ValidationInfo
    ) -> D:
        del info

        match value:
            case str():
                return cls.from_str(value)
            case D():
                return value
            case _:
                msg = f"Expected str or D, got {type(value)}"
                raise TypeError(msg)

    def __apply_shorthand_formatting(self) -> D:
        """Apply shorthand formatting based on the formatter settings."""
        formatter = serialize.get_current_formatter()
        line = formatter.path_data_shorthand_line_commands
        curve = formatter.path_data_shorthand_curve_commands

        d = self

        if "never" in (line, curve):
            d = d.resolve_shorthands(
                lines=line == "never", curves=curve == "never"
            )

        if "always" in (line, curve):
            d = d.apply_shorthands(
                lines=line == "always", curves=curve == "always"
            )

        return d

    def __serialize_commands(self) -> Generator[str]:
        formatter = serialize.get_current_formatter()

        d = self.__apply_shorthand_formatting()

        if formatter.path_data_coordinates == "relative":
            d = _relativize(d)

        for prev, command in utils.pairwise(d):
            implicit = (
                formatter.path_data_commands == "implicit"
                and _can_use_implicit_command(command, prev=prev)
            )

            match command:
                case MoveTo(end):
                    yield serialize.serialize_path_command(
                        end, char="M", implicit=implicit
                    )
                case LineTo(end):
                    yield serialize.serialize_path_command(
                        end, char="L", implicit=implicit
                    )
                case HorizontalLineTo(x):
                    yield serialize.serialize_path_command(
                        x, char="H", implicit=implicit
                    )
                case VerticalLineTo(y):
                    yield serialize.serialize_path_command(
                        y, char="V", implicit=implicit
                    )
                case QuadraticBezierTo(control, end):
                    yield serialize.serialize_path_command(
                        control, end, char="Q", implicit=implicit
                    )
                case SmoothQuadraticBezierTo(end):
                    yield serialize.serialize_path_command(
                        end, char="T", implicit=implicit
                    )
                case CubicBezierTo(control1, control2, end):
                    yield serialize.serialize_path_command(
                        control1,
                        control2,
                        end,
                        char="C",
                        implicit=implicit,
                    )
                case SmoothCubicBezierTo(control2, end):
                    yield serialize.serialize_path_command(
                        control2, end, char="S", implicit=implicit
                    )
                case ArcTo(radii, angle, large, sweep, end):
                    yield serialize.serialize_path_command(
                        radii,
                        angle,
                        large,
                        sweep,
                        end,
                        char="A",
                        implicit=implicit,
                    )
                case ClosePath():
                    yield serialize.serialize_path_command(
                        char="Z", implicit=implicit
                    )

    @override
    def serialize(self) -> str:
        return " ".join(self.__serialize_commands())

    def resolve_shorthands(
        self, *, lines: bool = True, curves: bool = True
    ) -> Self:
        """Resolve shorthand commands into their full-length equivalents.

        Args:
            lines: Whether to resolve shorthand line commands (`H`, `V`) into
            the full-length equivalent (`L`).
            curves: Whether to resolve shorthand curve commands (`S`, `T`) into
            the full-length equivalent (`C`, `Q`).

        Returns:
            A new `D` instance with the shorthand commands (`H`, `V`, `S`, `T`)
            replaced by their full-length equivalents (`L`, `C`, `Q`).

        Examples:
        >>> d = D.from_str("M 0,0 H 10")
        >>> d.resolve_shorthands()
        D(MoveTo(end=Point(x=0.0, y=0.0)), LineTo(end=Point(x=10.0, y=0.0)))

        """
        d = type(self)()

        for command in self:
            match command:
                case SmoothQuadraticBezierTo(end=end) if curves:
                    control = _quadratic_control(self, command)
                    d.quadratic_bezier_to(control, end)
                case SmoothCubicBezierTo(control2=control2, end=end) if (
                    curves
                ):
                    control1 = _cubic_control(self, command)
                    d.cubic_bezier_to(control1, control2, end)
                case HorizontalLineTo(x=x) if lines:
                    end = _get_end(d, d[-1])
                    d.line_to(point.Point(x, end.y))
                case VerticalLineTo(y=y) if lines:
                    end = _get_end(d, d[-1])
                    d.line_to(point.Point(end.x, y))
                case _:
                    d.append(command)
        return d

    def apply_shorthands(
        self, *, lines: bool = True, curves: bool = True
    ) -> Self:
        """Convert full-length commands into shorthand commands where possible.

        Args:
            lines: Whether to convert line commands (`L`) to shorthand
            (`H`, `V`) where possible.
            curves: Whether to convert curve commands (`Q`, `C`) to shorthand
            (`T`, `S`) where possible.

        Returns:
            A new `D` instance with the full-length commands (`L`, `C`, `Q`)
            replaced by their shorthand equivalents (`H`, `V`, `S`, `T`).

        Examples:
        >>> d = D.from_str("M 10,10 L 100,10")
        >>> d.apply_shorthands()
        D(MoveTo(end=Point(x=10.0, y=10.0)), HorizontalLineTo(x=100.0))

        """
        d = type(self)()

        for command in self:
            match command:
                case LineTo(end=end) if lines and end.x == (
                    _get_end(d, d[-1]).x
                ):
                    d.vertical_line_to(end.y)
                case LineTo(end=end) if lines and end.y == (
                    _get_end(d, d[-1]).y
                ):
                    d.horizontal_line_to(end.x)
                case QuadraticBezierTo(control=control, end=end) if curves:
                    d.smooth_quadratic_bezier_to(end)

                    shorthand = d[-1]
                    assert isinstance(shorthand, SmoothQuadraticBezierTo)

                    # try to replace the command with a shorthand
                    auto_control = _quadratic_control(d, shorthand)

                    # if the shorthand turns out not to be compatible, revert
                    # to the original command
                    if control != auto_control:
                        d[-1] = command

                case CubicBezierTo(
                    control1=control1, control2=control2, end=end
                ) if curves:
                    d.smooth_cubic_bezier_to(control2, end)

                    shorthand = d[-1]
                    assert isinstance(shorthand, SmoothCubicBezierTo)

                    auto_control = _cubic_control(d, shorthand)

                    if control1 != auto_control:
                        d[-1] = command

                case _:
                    d.append(command)

        return d

    @overload
    def __getitem__(self, index: SupportsIndex) -> PathCommand: ...

    @overload
    def __getitem__(self, index: slice) -> Self: ...

    @override
    def __getitem__(
        self, index: SupportsIndex | slice
    ) -> PathCommand | Self:
        if isinstance(index, SupportsIndex):
            return self.__commands[index]

        return type(self)(self.__commands[index])

    @overload
    def __setitem__(
        self, index: SupportsIndex, values: PathCommand
    ) -> None: ...

    @overload
    def __setitem__(
        self, index: slice, values: Iterable[PathCommand]
    ) -> None: ...

    @override
    def __setitem__(
        self,
        index: SupportsIndex | slice,
        values: PathCommand | Iterable[PathCommand],
    ) -> None:
        if isinstance(index, slice):
            raise NotImplementedError(
                "__setitem__ with a slice is not supported"
            )

        assert isinstance(values, PathCommand)

        if utils.is_first_index(self, index) and not isinstance(
            values, MoveTo
        ):
            raise errors.SvgPathMissingMoveToError

        self.__commands[index] = values

    @overload
    def __delitem__(self, index: SupportsIndex) -> None: ...

    @overload
    def __delitem__(self, index: slice) -> None: ...

    @override
    def __delitem__(self, index: SupportsIndex | slice) -> None:
        if isinstance(index, slice):
            raise NotImplementedError(
                "__delitem__ with a slice is not supported"
            )

        if (
            utils.is_first_index(self, index)
            and len(self) > 1
            and not isinstance(self[int(index) + 1], MoveTo)
        ):
            raise errors.SvgPathMissingMoveToError

        del self.__commands[index]

    @override
    def __len__(self) -> int:
        return len(self.__commands)

    @override
    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(
            other @ command
            if isinstance(command, _PhysicalPathCommand)
            else command
            for command in self
        )

    @override
    def __eq__(self, other: object) -> bool:
        if not utils.basic_compare(other, self=self):
            return False

        if len(self) != len(other):
            return False

        return all(c1 == c2 for c1, c2 in zip(self, other, strict=True))

    @override
    def __repr__(self) -> str:
        name = type(self).__name__
        commands = ", ".join(repr(command) for command in self)
        return f"{name}({commands})"


@lark.v_args(inline=True)
@parse.visit_tokens  # there are a few terminals we want to parse
class _Transformer(lark.Transformer[object, D]):
    point = point.Point
    NUMBER = float

    cubic_bezier = CubicBezierTo
    horizontal_line = HorizontalLineTo
    line = LineTo
    move = MoveTo
    quadratic_bezier = QuadraticBezierTo
    smooth_cubic_bezier = SmoothCubicBezierTo
    smooth_quadratic_bezier = SmoothQuadraticBezierTo
    vertical_line = VerticalLineTo
    z = ClosePath

    def FLAG(self, value: _Flag) -> bool:  # noqa: N802
        return value == "1"

    def arc(
        self,
        radii: point.Point,
        angle: float,
        large: bool,  # noqa: FBT001
        sweep: bool,  # noqa: FBT001
        end: point.Point,
    ) -> ArcTo:
        return ArcTo(
            radii=radii, angle=angle, large=large, sweep=sweep, end=end
        )

    @lark.v_args(inline=False)
    def path(
        self, args: list[PathCommand | lark.Tree[PathCommand | lark.Token]]
    ) -> D:
        d = D()

        for item in args:
            match item:
                # simple commands like `Z` require no further processing
                case _PathCommandBase() as command:
                    _add_command(d, command)
                # commands that are part of a group need to be extracted;
                # the relative flag is applied if the group is relative
                case lark.Tree(data=name, children=commands):
                    # sanity check for when the grammar changes
                    assert "relative" in name or "absolute" in name
                    relative = "relative" in name

                    for command in commands:
                        assert isinstance(command, PathCommand)
                        _add_command(d, command, relative=relative)

        return d


DType: TypeAlias = Annotated[
    D, parse.get_validator(grammar="d.lark", transformer=_Transformer())
]
